# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import asyncio
import datetime
import os
import re

from shc.base import Subscribable
from shc.timer import Every

WINDOWS = os.name == 'nt'


class Ping(Subscribable[bool]):
    """
    A simple *Subscribable* object that periodically checks a given network host to be alive, using the ping command,
    and publishes the result as a bool value.

    This can be used for a simple presence monitoring, e.g. by pinging your mobile phone in your local WiFi network.

    :param address: The network host to be checked. Can be any string that can be passed as the host argument to the
        `ping` command, i.e. an IPv4 address, an IPv6 address, a host name of a full-qualified domain name.
    :param interval: The interval in which the host state shall be checked by pinging it
    :param number: The number of ECHO requests to send each time. Passed to `ping` via the `-c` argument (resp. `-n` on
        Windows). The host is considered to be alive (= `True` is published) if any of the sent pings is successfully
        received.
    :param timeout: The timeout for each individual ECHO request in seconds, passed to `ping` via the `-W` argument
        (resp. `-w` on Windows)
    """
    type = bool
    RE_WIN_PING_TTL = re.compile(rb'TTL=\d+\s*\n')

    def __init__(self, address: str, interval: datetime.timedelta = datetime.timedelta(minutes=5), number: int = 5,
                 timeout: float = 1.0):
        super().__init__()
        self.address = address
        self.number = number
        self.timeout = timeout
        self._timer = Every(interval)
        self._timer.trigger(self._exec)

    async def _exec(self, _v, _o) -> None:
        ping_process = await asyncio.create_subprocess_exec(
            'ping',
            '-c' if not WINDOWS else '-n',
            str(self.number),
            '-W' if not WINDOWS else '-w',
            str(self.timeout) if not WINDOWS else str(round(self.timeout * 1000)),
            self.address,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )
        std_out, _std_err = await ping_process.communicate()
        if ping_process.returncode != 0:
            self._publish(False, [])
            return
        if WINDOWS and not self.RE_WIN_PING_TTL.search(std_out):
            self._publish(False, [])
            return
        self._publish(True, [])



# Copyright 2020-2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import asyncio
import contextvars
import functools
import logging
from typing import Generic, List, Any, Tuple, Callable, Optional, Type, TypeVar, Awaitable, Union, Dict, Set

from . import conversion

S = TypeVar('S')
T = TypeVar('T')
T_co = TypeVar('T_co')
T_con = TypeVar('T_con')
LogicHandler = Callable[[T, List[Any]], Awaitable[None]]

logger = logging.getLogger(__name__)

magicOriginVar: contextvars.ContextVar[List[Any]] = contextvars.ContextVar('shc_origin')

C = TypeVar('C', bound="Connectable")


class Connectable(Generic[T], metaclass=abc.ABCMeta):
    """
    :cvar type: The type of the values, this object is supposed to handle
    """
    type: Type[T]

    def connect(self: C, other: "Connectable[S]",
                send: Optional[bool] = None,
                receive: Optional[bool] = None,
                read: Optional[bool] = None,
                provide: Optional[bool] = None,
                convert: Union[bool, Tuple[Callable[[T], S], Callable[[S], T]]] = False) -> C:
        """
        Subscribe self to other and set as default_provider and vice versa (depending on the two object's capabilities,
        optionalities and given parameters).

        :param other: The other *Connectable* object to connect with
        :param send: Send value updates to `other` (i.e. subscribe `other` to `self`).
            Requires `self` to be :class:`Subscribable` and other to be :class:`Writable`. In this case, defaults to
            True if not specified.
        :param receive: Receive value updates from `other` (i.e. subscribe `self` to `other`).
            Requires `self` to be :class:`Writable` and other to be :class:`Subscribable`. In this case, defaults to
            True if not specified.
        :param read: Read values from `other` (i.e. set `other` as default provider for `self`).
            Requires `self` to be :class:`Reading` and `other` to be :class:`Readable`.
            If not specified, defaults to False unless `self.is_reading_optional` is set to False.
        :param provide: Provide values to `other` for reading (i.e. set `self` as default provider for `other`).
            Requires `self` to be :class:`Readable` and `other` to be :class:`Reading`.
            If not specified, defaults to False unless `other.is_reading_optional` is set to False.
        :param convert: Enable built-in type conversion for the created subscriptions/default_providers:
            Either a boolean or a tuple of two conversion functions. Defaults to ``False``, i.e. a type mismatch will
            result in a TypeError instead of implicit conversion. Set to ``True``, to choose the appropriate
            conversion function automatically. Raises a TypeError if no default conversion for one of the two directions
            is available and that direction is used.
            For custom conversion functions, pass a tuple ``(to_other, to_self)``, where ``to_other`` is a (non-async)
            callable for converting `self`'s value type into `other`'s value type and ``to_self`` does the conversion
            the other way round.
        :return: Returns self to allow functional-style chaining
        """
        if isinstance(other, ConnectableWrapper):
            # If other object is not connectable itself but wraps one or more connectable objects (like, for example, a
            # `web.widgets.ValueButtonGroup`), let it use its special implementation of `connect()`.
            other.connect(self, send=receive, receive=send, read=provide, provide=read,
                          convert=((convert[1], convert[0]) if isinstance(convert, tuple) else convert))
        else:
            self._connect_with(self, other, send, provide, convert[0] if isinstance(convert, tuple) else convert)
            self._connect_with(other, self, receive, read, convert[1] if isinstance(convert, tuple) else convert)
        return self

    @staticmethod
    def _connect_with(source: "Connectable", target: "Connectable", send: Optional[bool], provide: Optional[bool],
                      convert: Union[bool, Callable]):
        if isinstance(source, Subscribable) and isinstance(target, Writable) and (send or send is None):
            source.subscribe(target, convert=convert)
        elif send and not isinstance(source, Subscribable):
            raise TypeError("Cannot subscribe {} to {}, since the latter is not Subscribable".format(target, source))
        elif send and not isinstance(target, Writable):
            raise TypeError("Cannot subscribe {} to {}, since the former is not Writable".format(target, source))
        if isinstance(source, Readable) and isinstance(target, Reading) \
                and (provide or (provide is None and not target.is_reading_optional)):
            target.set_provider(source, convert=convert)
        elif provide and not isinstance(source, Readable):
            raise TypeError("Cannot use {} as read provider for {}, since the former is not Readable"
                            .format(source, target))
        elif provide and not isinstance(target, Reading):
            raise TypeError("Cannot use {} as read provider for {}, since the latter is not Reading"
                            .format(source, target))


class ConnectableWrapper(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
    type: Type[T]

    @abc.abstractmethod
    def connect(self: C, other: "Connectable",
                send: Optional[bool] = None,
                receive: Optional[bool] = None,
                read: Optional[bool] = None,
                provide: Optional[bool] = None,
                convert: Union[bool, Tuple[Callable[[T], Any], Callable[[Any], T]]] = False) -> C:
        pass


class Writable(Connectable[T_con], Generic[T_con], metaclass=abc.ABCMeta):
    async def write(self, value: T_con, origin: Optional[List[Any]] = None) -> None:
        """
        Asynchronous coroutine to update the object with a new value

        This method calls :meth:`_write` internally for the actual implementation-specific update logic. Inheriting
        classes should override *_write* instead of this method to keep profiting from the value type checking and
        magic context-based origin passing features.

        This method awaits the complete transmission and processing of the new value by the next stateful
        object/system, e.g. storing and re-publishing the value on a Variable, publishing the value from an MQTT
        broker on an MQTT client connector, forwarding the value to all subscribers on an Expression.

        :param value: The new value
        :param origin: The origin / trace of the value update event, i.e. the list of objects/functions which have been
            publishing to/calling one another to cause this value update. It is used to avoid recursive feedback loops
            and may be used for origin-specific handling of the value. The last entry of the list should be the
            object/function calling this method.
        :raises TypeError: If the value's type does not match the the object's ``type`` attribute.
        """
        if origin is None:
            try:
                origin = magicOriginVar.get()
            except LookupError as e:
                raise ValueError("No origin attribute provided or set via execution context") from e
        if not isinstance(value, self.type):
            raise TypeError("Invalid type for {}: {} is not a {}".format(self, value, self.type.__name__))
        logger.debug("New value %s for %s via %s", value, self, origin)
        await self._write(value, origin)

    @abc.abstractmethod
    async def _write(self, value: T_con, origin: List[Any]) -> None:
        """
        Abstract internal method containing the actual implementation-specific write-logic.

        It must be overridden by classes inheriting from :class:`Writable` to be updated with new values. The *_write*
        implementation does not need to check the new value's type.

        Please make sure that your `_write` implementation awaits the processing of the value update by the next
        stateful object/system before returning:

        - On a Variable and similar stateful re-publishing objects, which use `_stateful_publishing = True`, await the
          storing and `_publish()`-ing of the new value
        - On an Expression and similar stateless re-publishing objects, use `_publish_and_wait()` instead of
          `_publish()` and await its return
        - On external interface connectors (e.g. MQTT connector) await (at least) the successful processing of the value
          by the external system

        This is required to make the state inconsistency mitigation method of :class:`Subscribable` objects work.

        :param value: The new value to update this object with
        :param origin: The origin / trace of the value update event. Should be passed to :meth:`Subscribable._publish`
            if the implementing class is *Subscribable* and re-publishes new values.
        """
        pass


class Readable(Connectable[T_co], Generic[T_co], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def read(self) -> T_co:
        """
        Get the current value of this Connectable object.

        :raises UninitializedError: If the value is not (yet) specified
        """
        pass


class UninitializedError(RuntimeError):
    pass


SubscriberListType = Tuple[Writable[S], Optional[Callable[[T_co], S]]]


class Subscribable(Connectable[T_co], Generic[T_co], metaclass=abc.ABCMeta):
    _stateful_publishing: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subscribers: List[SubscriberListType] = []
        self._triggers: List[Tuple[LogicHandler, bool]] = []
        self._pending_updates: Dict[int, Dict[asyncio.Task, Optional[int]]] = {}

    async def __publish_write(self, subscriber: Writable[S], converter: Optional[Callable[[T_co], S]], value: T_co,
                              origin: List[Any], use_pending: bool):
        try:
            await subscriber.write(converter(value) if converter else value, origin + [self])  # type: ignore
        except Exception as e:
            logger.error("Error while writing new value %s from %s to %s:", value, self, subscriber, exc_info=e)
        finally:
            if use_pending:
                del self._pending_updates[id(subscriber)][asyncio.current_task()]  # type: ignore

    async def __publish_trigger(self, target: LogicHandler, value: T_co,
                                origin: List[Any], use_pending: bool):
        try:
            await target(value, origin + [self])
        except Exception as e:
            logger.error("Error while triggering %s from %s:", target, self, exc_info=e)
        finally:
            if use_pending:
                del self._pending_updates[id(target)][asyncio.current_task()]  # type: ignore

    def _publish(self, value: T_co, origin: List[Any]):
        """
        Method to publish a new value to all subscribers and trigger all registered logic handlers.

        All logic handlers and :meth:`Writable.write` methods are called in parallel asyncio tasks, which are **not**
        awaited to return. If `_stateful_publishing` is True on this object (or class), it will keep track of the
        currently pending publishing tasks. This information is then used to mitigate state inconsistencies, caused by
        value updates crossing over each other, by resetting the `origin` of following value updates to the respected
        subscribers.

        `_stateful_publishing` should be enabled for all objects that are *Subscribable* and *Writable* (i.e. allows
        sending and receiving value updates) and have either internal state (like Variables) or represent an external,
        stateful (or quasi-stateful) system (like a Connector object for a KNX group address/an MQTT topic). On the
        other hand, objects, that are only *Subscribable*, should not use `_stateful_publishing`.

        If the object only forwards value updates without storing a state, it should await the successful publishing of
        the value by using the :meth:`_publish_and_wait` coroutine.

        :param value: The new value to be published by this object. Must be an instance of this object's `type`.
        :param origin: The origin list of the new value, **excluding** this object. See :ref:`base.event-origin` for
            more details.`self` is appended automatically before calling the registered subscribers and logic handlers.
        """
        if self._stateful_publishing:
            for subscriber, converter in self._subscribers:
                prev_step = id(origin[-1]) if origin else None
                reset_origin = any(o != prev_step for o in self._pending_updates[id(subscriber)].values())
                if reset_origin:
                    logger.info("Resetting origin from %s to %s; value=%s; origin=%s", self, subscriber, value, origin)
                if reset_origin or not any(s is subscriber for s in origin):
                    task = asyncio.create_task(self.__publish_write(subscriber, converter, value,
                                                                    [] if reset_origin else origin, True))
                    self._pending_updates[id(subscriber)][task] = prev_step
            for target, sync in self._triggers:
                reset_origin = False
                if sync:
                    prev_step = id(origin[-1]) if origin else None
                    reset_origin = any(o != prev_step for o in self._pending_updates[id(target)].values())
                task = asyncio.create_task(self.__publish_trigger(target, value, [] if reset_origin else origin, sync))
                if sync:
                    self._pending_updates[id(target)][task] = prev_step

        else:
            for target, sync in self._triggers:
                asyncio.create_task(self.__publish_trigger(target, value, origin, False))
            for subscriber, converter in self._subscribers:
                if not any(s is subscriber for s in origin):
                    asyncio.create_task(self.__publish_write(subscriber, converter, value, origin, False))

    async def _publish_and_wait(self, value: T_co, origin: List[Any]):
        """
        A coroutine to publish a new value to all subscribers and trigger all registered logic handlers and wait for
        this publishing to finish.

        All logic handlers and :meth:`Writable.write` methods are called in parallel asyncio tasks. However, this
        method awaits the return of *all* them. This should only be used for **stateless** objects that re-publish
        incoming value updates (like :ref:`Expressions <expressions>`, :class:`shc.misc.TwoWayPipe`, etc.). Other
        objects should use :meth:`_publish` instead.

        :param value: The new value to be published by this object. Must be an instance of this object's `type`.
        :param origin: The origin list of the new value, **excluding** this object. See :ref:`base.event-origin` for
            more details.`self` is appended automatically before calling the registered subscribers and logic handlers.
        """
        for target, sync in self._triggers:
            if not sync:
                asyncio.create_task(self.__publish_trigger(target, value, origin, False))
        sync_jobs = [self.__publish_write(subscriber, converter, value, origin, False)
                     for subscriber, converter in self._subscribers
                     if not any(s is subscriber for s in origin)]
        sync_jobs.extend(
            self.__publish_trigger(target, value, origin, False)
            for target, sync in self._triggers
            if sync)
        if len(sync_jobs) == 1:
            await sync_jobs[0]
        elif sync_jobs:
            await asyncio.gather(*sync_jobs)

    def subscribe(self, subscriber: Writable[S], convert: Union[Callable[[T_co], S], bool] = False) -> None:
        """
        Subscribe a writable object to this object to be updated, when this object publishes a new value.

        The subscriber's :meth:`Writable.write` method will be called for any new value published by this object, as
        long as the subscriber did not lead to the relevant update of this object (i.e. is not included in the
        `origin` list). The origin list passed to the subscriber's `write` method will contain this object as the last
        entry.

        :param subscriber: The object to subscribe for updates
        :param convert: A callable to convert this object's new value to the data ``type`` of the subscriber or ``True``
            to choose the appropriate conversion function automatically.
        :raises TypeError: If the `type` of the subscriber does not match this object's type and ``convert`` is False
            *or* if ``convert`` is True but no type conversion is known to convert this object's type into the
            subscriber's type.
        """
        converter: Optional[Callable[[T_co], S]]
        if callable(convert):
            converter = convert
        elif issubclass(self.type, subscriber.type):
            converter = None
        elif convert:
            converter = conversion.get_converter(self.type, subscriber.type)
        else:
            raise TypeError("Type mismatch of subscriber {} ({}) for {} ({})"
                            .format(repr(subscriber), subscriber.type.__name__, repr(self), self.type.__name__))
        self._subscribers.append((subscriber, converter))
        if self._stateful_publishing:
            self._pending_updates[id(subscriber)] = {}

    def trigger(self, target: LogicHandler, synchronous: bool = False) -> LogicHandler:
        """
        Register a logic handler function to be triggered when this object is updated.

        This method can be used as a decorator for custom logic handler functions. Alternatively, it can simply called
        with a function object: ``some_subscribable.trigger(some_handler_function)``.

        The `target` function must be an async coroutine that takes two arguments: The new value of this object and the
        origin/trace of the event (a list of objects that led to the handler being tiggered). The handler function
        must make sure to prevent infinite recursive feedback loops: In contrast to subscribed objects, logic handler
        functions are also triggered, if they led to the object being updated (i.e. they are already conained in the
        ``origin`` list). Thus, they should skip execution if called recursively. It should also append itself to the
        ``origin`` list and pass the extended list to all :meth:`Writable.write` calls it does.

        To ensure all this for a custom handler function, use the :func:`handler` decorator::

            @some_subscribable.trigger
            @handler()
            async def some_handler_function(value, origin):
                ...
                some_writable.write(value + 1)
                ...

        You may even use multiple trigger decorators::

            @some_subscribable.trigger
            @another_subscribable.trigger
            @handler()
            async def some_handler_function(value, origin):
                if origin[-1] is some_subscribable:
                    ...
                else:
                    ...

        :param target: The handler function/coroutine to be triggered on updates. Must comply with the requirements
            mentioned above.
        :param synchronous: If True, the target coroutine is triggered synchronously, i.e. :meth:`_publish_and_wait`
            will wait for it to return. This SHALL be used for internal triggering methods that re-publish the value
            updates of this Subscribable object. For logic handlers that might require substantial time to return, it
            SHALL not be used!
        :return: The ``target`` function (unchanged) – this allows decorator-style usage of this method
        """
        self._triggers.append((target, synchronous))
        if synchronous and self._stateful_publishing:
            self._pending_updates[id(target)] = {}
        return target


DefaultProviderType = Tuple[Readable[S], Optional[Callable[[S], T_con]]]


class Reading(Connectable[T_con], Generic[T_con], metaclass=abc.ABCMeta):
    is_reading_optional: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._default_provider: Optional[DefaultProviderType] = None

    def set_provider(self, provider: Readable[S], convert: Union[Callable[[S], T_con], bool] = False):
        converter: Optional[Callable[[S], T_con]]
        if callable(convert):
            converter = convert
        elif issubclass(provider.type, self.type):
            converter = None
        elif convert:
            converter = conversion.get_converter(provider.type, self.type)
        else:
            raise TypeError("Type mismatch of Readable {} ({}) as provider for {} ({})"
                            .format(repr(provider), provider.type.__name__, repr(self), self.type.__name__))
        self._default_provider = (provider, converter)

    async def _from_provider(self) -> Optional[T_con]:
        """
        Private method to be used by inheriting classes to read the current value from this object's default provider
        (via its :meth:`Readable.read` method) and convert it to this object's type, if necessary, using the registered
        converter function.

        :return: The default provider's current value *or* ``None`` if no default provider is set or it's *read* method
            raises an :class:`UninitializedError`.
        """
        if self._default_provider is None:
            return None
        provider, convert = self._default_provider
        try:
            val = await provider.read()
        except UninitializedError:
            return None
        return convert(val) if convert else val


def handler(reset_origin=False, allow_recursion=False) -> Callable[[LogicHandler], LogicHandler]:
    """
    Decorator for custom logic handler functions.

    Wraps a custom logic handler functions to make sure it is suited to be registered for triggering by a subscribable
    object with :meth:`Subscribable.trigger`. It makes sure that

    * exceptions, occurring execution, are logged,
    * the `origin` is extended with the logic handler itself and magically passed to all :meth:`Writable.write` calls
    * the `origin` can magically be passed when called directly by other logic handlers
    * the execution is skipped when called recursively (i.e. the logic handler is already contained in the `origin` list

    :param reset_origin: If True, the origin which is magically passed to all `write` calls, only contains the logic
        handler itself, not the previous `origin` list, which led to the handler's execution. This can be used to
        change an object's value, which triggered this logic handler. This may cause infinite recursive feedback loops,
        so use with care!
    :param allow_recursion: If True, recursive execution of the handler is not skipped. The handler must check the
        passed values and/or the `origin` list itself to prevent infinite feedback loops via `write` calls or calls to
        other logic handlers – especiaally when used together with `reset_origin`.
    """
    def decorator(f: LogicHandler) -> LogicHandler:
        @functools.wraps(f)
        async def wrapper(value, origin: Optional[List[Any]] = None) -> None:
            if origin is None:
                try:
                    origin = magicOriginVar.get()
                except LookupError as e:
                    raise ValueError("No origin attribute provided or set via execution context") from e
            if any(s is wrapper for s in origin) and not allow_recursion:
                logger.info("Skipping recursive execution of logic handler %s() via %s", f.__name__, origin)
                return
            logger.info("Triggering logic handler %s() from %s", f.__name__, origin)
            try:
                token = magicOriginVar.set([wrapper] if reset_origin else (origin + [wrapper]))
                await f(value, origin)
                magicOriginVar.reset(token)
            except Exception as e:
                logger.error("Error while executing handler %s():", f.__name__, exc_info=e)
        return wrapper
    return decorator


def blocking_handler() -> Callable[[Callable[[T, List[Any]], None]], LogicHandler]:
    """
    Decorator for custom blocking (non-async) logic handler functions.

    Wraps a function to transform it into an async logic handler function, which is suited to be registered for
    triggering by a subscribable object with :meth:`Subscribable.trigger`. The wrapped function is executed in a
    separate thread, using asyncio's `run_in_executor()`.

    Like :func:`handler`, this decorator catches and logs errors and ensures that the `origin` can magically be passed
    when called directly by other logic handlers. However, since the wrapped function is not an asynchronous coroutine,
    it is not able to call :meth:`Writable.write` or another logic handler directly. Thus, this decorator does not
    include special measures for preparing and passing the `origin` list or avoiding recursive execution.
    """
    def decorator(f: Callable[[T, List[Any]], None]) -> LogicHandler:
        @functools.wraps(f)
        async def wrapper(value, origin: Optional[List[Any]] = None) -> None:
            if origin is None:
                try:
                    origin = magicOriginVar.get()
                except LookupError as e:
                    raise ValueError("No origin attribute provided or set via execution context") from e
            logger.info("Triggering blocking logic handler %s() from %s", f.__name__, origin)
            try:
                loop = asyncio.get_event_loop()
                await loop.run_in_executor(None, f, value, origin)
            except Exception as e:
                logger.error("Error while executing handler %s():", f.__name__, exc_info=e)
        return wrapper
    return decorator



# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""
A MyPy type checker plugin for some of the dynamic typing features of the SmartHomeConnect framework (SHC)

Currently, it allows to improve the typechecking for VariableField objects which are retrieved via the
:meth:`field() <shc.variables.Variable.field>` method of `Variables` and `VariableFields`.

To enable this plugin, set `plugins = shc.util.mypy_variable_plugin` in the global `[mypy]` section of your MyPy config
file (other Plugins can be added with comma-separation).
"""
from typing import Optional, Callable

from mypy import errorcodes
from mypy.nodes import StrExpr
from mypy.plugin import Plugin, MethodContext
from mypy.types import Instance, Type, TupleType


class SHCVariable(Plugin):
    def get_method_hook(self, fullname: str) -> Optional[Callable[[MethodContext], Type]]:
        if fullname in ("shc.variables.Variable.field", "shc.variables.VariableField.field"):
            return self._shc_variable_field_callback
        return None

    @staticmethod
    def _shc_variable_field_callback(context: MethodContext) -> Type:
        assert isinstance(context.type, Instance)
        assert isinstance(context.default_return_type, Instance)
        if not context.type.args:
            return context.default_return_type
        wrapped_type = context.type.args[0]
        if not isinstance(wrapped_type, TupleType) or not wrapped_type.partial_fallback.type.is_named_tuple:
            context.api.fail(".field() can only be used with NamedTuple-typed SHC Variables or VariableFields",
                             context.context, code=errorcodes.MISC)
            return context.default_return_type
        field_name_expression = context.args[0][0]
        if not isinstance(field_name_expression, StrExpr):
            # If the field name is not specified as a simple string expression, we cannot check the value
            return context.default_return_type
        field_name = field_name_expression.value
        tuple_field_names = wrapped_type.partial_fallback.type.metadata['namedtuple']['fields']
        assert isinstance(tuple_field_names, list)
        try:
            field_index = tuple_field_names.index(field_name)
        except ValueError:
            context.api.fail(f"NamedTuple \"{wrapped_type}\" has no field \"{field_name}\" to create a variable field "
                             f"for",
                             context.context, code=errorcodes.ATTR_DEFINED)
            return context.default_return_type
        field_type = wrapped_type.items[field_index]
        return Instance(context.default_return_type.type, [field_type])


def plugin(version: str):
    return SHCVariable



# Copyright 2020-2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
import asyncio
import datetime
import functools
import math
import operator
import typing
from typing import Type, Generic, Any, Iterable, Callable, Union, Dict, Tuple, List, Optional

from . import conversion
from .base import Readable, Subscribable, T, Connectable, Writable, S, LogicHandler, UninitializedError
from .datatypes import RangeFloat1, RangeUInt8, HSVFloat1, RGBFloat1, RGBUInt8


class ExpressionBuilder(Connectable[T], metaclass=abc.ABCMeta):
    """
    An abstract base class for any object that may be used in SHC expressions.

    This class defines all the overloaded operators, to create *Connectable* :class:`ExpressionHandler` objects when
    being used in a Python expression. For using this class's features with any *Readable* + *Subscribable* object, use
    the :class:`ExpressionWrapper` subclass. Additionally, each :class:`ExpressionHandler` is also an
    `ExpressionBuilder`, allowing to combine Expressions to even larger Expressions.

    Currently, the following operators and builtin functions are supported:

    * ``+`` (add)
    * ``-`` (sub)
    * ``*`` (mul)
    * ``/`` (truediv)
    * ``//`` (floordiv)
    * ``%`` (mod)
    * ``abs()``
    * ``-()`` (neg)
    * ``ceil()``
    * ``floor()``
    * ``round()``
    * ``==`` (eq)
    * ``!=`` (ne)
    * ``<`` (lt)
    * ``<=`` (le)
    * ``>`` (gt)
    * ``>=`` (ge)

    Additionally, the following methods are provided:
    """
    @staticmethod
    def __get_other_type(other: object) -> type:
        if isinstance(other, Readable) and isinstance(other, Subscribable):
            return other.type
        else:
            return type(other)

    def __add__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_ADD:
            return BinaryExpressionHandler(TYPES_ADD[(self.type, other_type)], self, other, operator.add)
        else:
            return NotImplemented

    def __radd__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_ADD:
            return BinaryExpressionHandler(TYPES_ADD[(other_type, self.type)], other, self, operator.add)
        else:
            return NotImplemented

    def __sub__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_SUB:
            return BinaryExpressionHandler(TYPES_SUB[(self.type, other_type)], self, other, operator.sub)
        else:
            return NotImplemented

    def __rsub__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_SUB:
            return BinaryExpressionHandler(TYPES_SUB[(other_type, self.type)], other, self, operator.sub)
        else:
            return NotImplemented

    def __mul__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_MUL:
            return BinaryExpressionHandler(TYPES_MUL[(self.type, other_type)], self, other, operator.mul)
        else:
            return NotImplemented

    def __rmul__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_MUL:
            return BinaryExpressionHandler(TYPES_MUL[(other_type, self.type)], other, self, operator.mul)
        else:
            return NotImplemented

    def __truediv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_TRUEDIV:
            return BinaryExpressionHandler(TYPES_TRUEDIV[(self.type, other_type)], self, other, operator.truediv)
        else:
            return NotImplemented

    def __rtruediv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_TRUEDIV:
            return BinaryExpressionHandler(TYPES_TRUEDIV[(other_type, self.type)], other, self, operator.truediv)
        else:
            return NotImplemented

    def __floordiv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_FLOORDIV:
            return BinaryExpressionHandler(TYPES_FLOORDIV[(self.type, other_type)], self, other, operator.floordiv)
        else:
            return NotImplemented

    def __rfloordiv__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_FLOORDIV:
            return BinaryExpressionHandler(TYPES_FLOORDIV[(other_type, self.type)], other, self, operator.floordiv)
        else:
            return NotImplemented

    def __mod__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_MOD:
            return BinaryExpressionHandler(TYPES_MOD[(self.type, other_type)], self, other, operator.floordiv)
        else:
            return NotImplemented

    def __rmod__(self, other) -> "BinaryExpressionHandler":
        other_type = self.__get_other_type(other)
        if (other_type, self.type) in TYPES_MOD:
            return BinaryExpressionHandler(TYPES_MOD[(other_type, self.type)], other, self, operator.mod)
        else:
            return NotImplemented

    def __abs__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_ABS_NEG:
            return UnaryExpressionHandler(TYPES_ABS_NEG[self.type], self, operator.abs)
        else:
            raise TypeError("{} cannot be used with abs()".format(self))

    def __neg__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_ABS_NEG:
            return UnaryExpressionHandler(TYPES_ABS_NEG[self.type], self, operator.neg)
        else:
            raise TypeError("{} cannot be negated".format(self))

    def __ceil__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, math.ceil)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __floor__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, math.floor)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __round__(self) -> "UnaryExpressionHandler":
        if self.type in TYPES_CEIL_FLOOR_ROUND:
            return UnaryExpressionHandler(TYPES_CEIL_FLOOR_ROUND[self.type], self, round)
        else:
            raise TypeError("{} cannot be rounded".format(self))

    def __eq__(self, other) -> "BinaryExpressionHandler[bool]":  # type: ignore
        return BinaryExpressionHandler(bool, self, other, operator.eq)

    def __ne__(self, other) -> "BinaryExpressionHandler[bool]":  # type: ignore
        return BinaryExpressionHandler(bool, self, other, operator.eq)

    def __lt__(self, other) -> "BinaryExpressionHandler[bool]":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.lt)
        else:
            return NotImplemented

    def __le__(self, other) -> "BinaryExpressionHandler[bool]":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.le)
        else:
            return NotImplemented

    def __gt__(self, other) -> "BinaryExpressionHandler[bool]":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.gt)
        else:
            return NotImplemented

    def __ge__(self, other) -> "BinaryExpressionHandler[bool]":
        other_type = self.__get_other_type(other)
        if (self.type, other_type) in TYPES_LT_LE_GT_GE:
            return BinaryExpressionHandler(TYPES_LT_LE_GT_GE[(self.type, other_type)], self, other, operator.ge)
        else:
            return NotImplemented

    def and_(self, other) -> "BinaryCastExpressionHandler":
        return BinaryCastExpressionHandler(bool, self, other, operator.and_)

    def or_(self, other) -> "BinaryCastExpressionHandler":
        return BinaryCastExpressionHandler(bool, self, other, operator.or_)

    def not_(self) -> "UnaryExpressionHandler[bool]":
        return UnaryCastExpressionHandler(bool, self, operator.not_)

    def convert(self, type_: Type[S], converter: Optional[Callable[[T], S]] = None) -> "UnaryExpressionHandler[S]":
        """
        Returns an ExpressionHandler that wraps this object to convert values to another type using a given converter
        function or the :ref:`default converter <datatypes.default_converters>` for those two types.

        Example::

            value = shc.Variable(float)
            which_one = shc.Variable(bool)
            # display_string shall be a str-typed expression object
            display_string = expression.IfThenElse(which_one, value.EX.convert(str), "static string")

        :param type_: The target type to convert new values to. This will also by the `type` attribute of the returned
            ExpressionHandler object, such that it will be used for further
        :param converter: An optional converter function to use instead of the default conversion from this object's
            type to the given target type.
        """
        converter = converter or conversion.get_converter(self.type, type_)
        return UnaryExpressionHandler(type_, self, converter)


class ExpressionWrapper(Readable[T], Subscribable[T], ExpressionBuilder, Generic[T]):
    """
    Wrapper for any *Readable* + *Subscribable* object to equip it with expression-building capabilities, inherited from
    :class:`ExpressionBuilder`
    """
    def __init__(self, wrapped: Subscribable[T]):
        super().__init__()
        self.wrapped = wrapped
        self.type = wrapped.type

    async def read(self) -> T:
        return await self.wrapped.read()  # type: ignore

    def subscribe(self, subscriber: Writable[S], convert: Union[Callable[[T], S], bool] = False):
        return self.wrapped.subscribe(subscriber, convert)

    def trigger(self, target: LogicHandler, synchronous: bool = False) -> LogicHandler:
        return self.wrapped.trigger(target, synchronous=synchronous)


class ExpressionHandler(Readable[T], Subscribable[T], ExpressionBuilder, Generic[T], metaclass=abc.ABCMeta):
    """
    Base class for expression objects, i.e. the result object of an expression built with :class:`ExpressionBuilder`'s
    overloaded operators

    The ExpressionHandler object stores the operator (as a callable) and a reference to each operand. Thereby, it can
    *read* there current values and evaluate the operation/expression at any time with these values. *ExpressionHandler*
    objects are *Readable* (to evaluate the expression's current value on demand) and *Subscribable* (to evaluate and
    publish the expression's new value, when any of the operands is updated).
    """
    def __init__(self, type_: Type[T], operands: Iterable[object]):
        super().__init__()
        self.type = type_
        for operand in operands:
            if isinstance(operand, Subscribable):
                operand.trigger(self.on_change, synchronous=True)

    @abc.abstractmethod
    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        pass

    async def on_change(self, value, origin):
        try:
            await self._publish_and_wait(await self.evaluate(value, origin[-1]), origin)
        except UninitializedError:
            pass

    async def read(self) -> T:
        return await self.evaluate(None, None)

    @staticmethod
    def _wrap_static_value(val: Union[S, Readable[S]]) -> Readable[S]:
        class Wrapper(Readable[S]):
            def __init__(self, v: S):
                self.v = v
                self.type = type(v)

            async def read(self) -> S:
                return self.v

        if isinstance(val, Readable):
            return val
        else:
            return Wrapper(val)


# ################################# #
# Operator-based ExpressionHandlers #
# ################################# #

class BinaryExpressionHandler(ExpressionHandler[T], Generic[T]):
    def __init__(self, type_: Type[T], a, b, operator_: Callable[[Any, Any], T]):
        self.a = self._wrap_static_value(a)
        self.b = self._wrap_static_value(b)
        self.operator = operator_
        super().__init__(type_, (a, b))

    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        return self.operator(received_value if received_from is self.a else await self.a.read(),
                             received_value if received_from is self.b else await self.b.read())

    def __repr__(self) -> str:
        return "{}[{}({}, {})]".format(self.__class__.__name__, self.operator.__name__, repr(self.a), repr(self.b))


class BinaryCastExpressionHandler(BinaryExpressionHandler[T]):
    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        return self.type(await super().evaluate(received_value, received_from))  # type: ignore


class UnaryExpressionHandler(ExpressionHandler[T], Generic[T]):
    def __init__(self, type_: Type[T], a, operator_: Callable[[Any], T]):
        self.a = self._wrap_static_value(a)
        self.operator = operator_
        super().__init__(type_, (a,))

    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        return self.operator(received_value if received_from is self.a else await self.a.read())

    def __repr__(self) -> str:
        return "{}[{}({})]".format(self.__class__.__name__, self.operator.__name__, repr(self.a))


class UnaryCastExpressionHandler(UnaryExpressionHandler[T]):
    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        return self.type(await super().evaluate(received_value, received_from))  # type: ignore


# ############################# #
# Functional ExpressionHandlers #
# ############################# #

class IfThenElse(ExpressionHandler, Generic[T]):
    """
    A :class:`ExpressionHandler` version of the ``x if condition else y`` python syntax

    This class takes three connectable objects or static values: `condition`, `then` and `otherwise`. `condition` must
    be ``bool`` (or a *Connectable* with value type ``bool``), `then` and `otherwise` must be of the same type (reps.
    have the same value type). If the `condition` evaluates to True, this object evaluates to the value of `then`,
    otherwise it evaluates to the value of `otherwise`.

    See also :class:`Multiplexer` for switching between more than two values using an integer control value.
    If you only want enable and disable a single subscription dynamically (e.g. enable/disable the influence of a
    :class:`TimerSwitch <shc.timer.TimerSwitch>` on a :class:`Variable <shc.variables.Variable>`), take a look at
    :class:`shc.misc.BreakableSubscription` instead.
    """
    def __init__(self, condition: Union[bool, Readable[bool]], then: Union[T, Readable[T]],
                 otherwise: Union[T, Readable[T]]):
        super().__init__(then.type if isinstance(then, Connectable) else type(then), (condition, then, otherwise))
        self.condition: Readable[bool] = self._wrap_static_value(condition)
        self.then = self._wrap_static_value(then)
        self.otherwise = self._wrap_static_value(otherwise)

    async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
        return ((received_value if received_from is self.then else await self.then.read())  # type: ignore
                if (received_value if received_from is self.condition else await self.condition.read())
                else (received_value if received_from is self.otherwise else await self.otherwise.read()))


class Multiplexer(Readable[T], Subscribable[T], ExpressionBuilder[T], Generic[T]):
    """
    A :class:`ExpressionHandler` that behaves as a multiplexer block with an integer control value and any number of
    inputs from that the output value is chosen

    The control object, the first argument of this class, needs to be a Readable (and optionally Subscribable)
    object of int type. All other parameters are considered to be inputs. They must either be static values or
    Readable (and optionally Subscribable) objects. All of them must be of the same value type (resp. be instances of
    exactly that type). This type will also be the type of the Multiplexer itself.

    The control signal selects which of the inputs is passed through and is the provided value of the Multiplexer. E.g.,
    when the control object has the value 0, *reading* from the Multiplexer returns the first input's current value
    (i.e. the value of the object given as the second argument); when the control object has the value 1, the second
    input's value is returned, and so on. If the control value is out of range of the available inputs, *reading*
    returns an :class:`UninitializedError`.

    The Multiplexer publishes a value update whenever an update from the control object or the currently selected
    input is received – as long as they are Subscribable.

    See also :class:`IfThenElse` for simple cases with only two values and a boolean control value.
    If you only want enable and disable a single subscription dynamically (e.g. enable/disable the influence of a
    :class:`TimerSwitch <shc.timer.TimerSwitch>` on a :class:`Variable <shc.variables.Variable>`), take a look at
    :class:`shc.misc.BreakableSubscription` instead.
    """
    _synchronous_publishing = True

    def __init__(self, control: Readable[int], *inputs: Union[T, Readable[T]]):
        super().__init__()
        if not inputs:
            raise ValueError("Multiplexer must have at least one input")
        self.inputs: List[Readable[T]] = [input_ if isinstance(input_, Readable)
                                          else ExpressionHandler._wrap_static_value(input_)
                                          for input_ in inputs]
        self.type: Type[T] = self.inputs[0].type
        for i, input_ in enumerate(self.inputs):
            if input_.type != self.type:
                raise ValueError("Type of input {} ({}) does not match: {} instead of {}"
                                 .format(i, input, input_.type, self.type))
            if isinstance(input_, Subscribable):
                input_.trigger(self._new_value, synchronous=True)

        self.control = control
        if isinstance(control, Subscribable):
            control.trigger(self._index_change, synchronous=True)

    async def read(self) -> T:
        try:
            current_index = await self.control.read()
            current_input = self.inputs[current_index]
        except IndexError:
            # We "ignore" index errors here. They may have been logged by _index_change() before
            raise UninitializedError()
        return await current_input.read()

    async def _new_value(self, value: T, origin: List[Any]) -> None:
        try:
            current_index = await self.control.read()
            current_input = self.inputs[current_index]
        except UninitializedError:
            # We default to False, if the control variable is not initialized yet
            return
        except IndexError:
            # We ignore index errors here. They may have been logged by _index_change() before
            return
        if current_input is origin[-1]:
            await self._publish_and_wait(value, origin)

    async def _index_change(self, current_index: int, origin: List[Any]) -> None:
        try:
            current_input = self.inputs[current_index]
        except IndexError:
            raise ValueError("Index {} is out of range of available inputs of {}".format(current_index, self))
        try:
            await self._publish_and_wait(await current_input.read(), origin)
        except UninitializedError:
            pass


def not_(a) -> Union[bool, UnaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps a *Connectable* or static value and evaluates to ``not x`` for the
    value `x`.

    This is the workaround for Python's limitations on overriding the ``not`` operator.
    """
    if isinstance(a, Readable) and isinstance(a, Subscribable):
        return UnaryCastExpressionHandler(bool, a, operator.not_)
    return not a


def and_(a, b) -> Union[bool, BinaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps two *Connectables* or static values a,b and evaluates to
    ``a and b``.

    This is the workaround for Python's limitations on overriding the ``and`` operator.
    """
    if (not (isinstance(a, Readable) and isinstance(a, Subscribable))
            and not (isinstance(b, Readable) and isinstance(b, Subscribable))):
        return bool(a and b)
    return BinaryCastExpressionHandler(bool, a, b, operator.and_)


def or_(a, b) -> Union[bool, BinaryCastExpressionHandler[bool]]:
    """
    Create a :class:`ExpressionHandler` that wraps two *Connectables* or static values a,b and evaluates to
    ``a or b``.

    This is the workaround for Python's limitations on overriding the ``or`` operator.
    """
    if (not (isinstance(a, Readable) and isinstance(a, Subscribable))
            and not (isinstance(b, Readable) and isinstance(b, Subscribable))):
        return bool(a or b)
    return BinaryCastExpressionHandler(bool, a, b, operator.or_)


# This is just a hack to make type checkers happy with the signature of the __init__ method of return types by
# expression
class ExpressionFunctionHandler(ExpressionHandler[T], Generic[T]):
    def __init__(self, *args: Any) -> None: ...


def expression(func: Callable[..., T]) -> Type[ExpressionFunctionHandler[T]]:
    """
    A decorator to turn any simple function into an :class:`ExpressionHandler` class to be used in SHC expressions

    This way, custom functions can be used in SHC expression in an intuitive way::

        @expression
        def invert(a: float) -> float:
            return 1 / a

        var1 = Variable(float)
        var2 = Variable(float).connect( invert(var1) )

    The decorator turns the function into a class, such that calling it will return an instance of that class instead of
    evaluating the function. The resulting object is a *Readable* and *Subscribable* object, that evaluates the
    original function with the current values of the given arguments on every *read()* and every value update. The
    arguments passed to each instance must either be *Readable* (and optionally *Subscribable*) objects, providing
    values of the expected argument type, or simply static objects of the expected type.

    The wrapped function must be synchronous (not async) and have a type annotation for its return type. The type
    annotation must refer to a runtime type, which can be used for dynamic type checking (i.e. no subscripted generic
    type; use `list` instead of `List[int]` and similar). This return type annotation is also the type of the resulting
    ExpressionHandler objects.

    Currently only positional arguments (no keyword arguments) are supported for the functions.
    """
    class WrapHandler(ExpressionHandler[T]):
        def __init__(self, *args):
            self.args = [self._wrap_static_value(a) for a in args]
            type_ = typing.get_type_hints(func).get('return')
            if not type_:
                raise TypeError("@expression requires the wrapped function to have a return type annotation")
            # Check if the type annotation is valid for runtime type checking
            try:
                isinstance(None, type_)
            except TypeError as e:
                raise TypeError("@expression requires the wrapped function to have a return type annotation") from e

            super().__init__(type_, args)

        async def _get_arg_value(self, arg, received_value, received_from):
            return received_value if received_from is arg else await arg.read()

        async def evaluate(self, received_value: Optional[Any], received_from: Optional[Any]) -> T:
            return func(
                *(await asyncio.gather(
                    *(self._get_arg_value(a, received_value, received_from)
                      for a in self.args))))

        def __repr__(self) -> str:
            return "@expression[{}({})]".format(func.__name__, repr(self.args))

    functools.update_wrapper(WrapHandler, func, updated=())
    return WrapHandler  # type: ignore


# #################### #
# Type Inference Rules #
# #################### #

TYPES_ADD: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (str, str): str,
    (bytes, bytes): bytes,
    (datetime.date, datetime.timedelta): datetime.date,
    (datetime.datetime, datetime.timedelta): datetime.datetime,
    (datetime.timedelta, datetime.date): datetime.date,
    (datetime.timedelta, datetime.datetime): datetime.datetime,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_SUB: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.date, datetime.date): datetime.timedelta,
    (datetime.datetime, datetime.datetime): datetime.timedelta,
    (datetime.date, datetime.timedelta): datetime.date,
    (datetime.datetime, datetime.timedelta): datetime.datetime,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_MUL: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (int, float): float,
    (float, int): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (str, int): str,
    (int, str): str,
    (bytes, int): bytes,
    (int, bytes): bytes,
    (datetime.timedelta, float): datetime.timedelta,
    (float, datetime.timedelta): datetime.timedelta,
    (datetime.timedelta, int): datetime.timedelta,
    (int, datetime.timedelta): datetime.timedelta,
    (RangeFloat1, float): float,
    (float, RangeFloat1): float,
    (RangeFloat1, int): int,
    (int, RangeFloat1): int,
    (RangeFloat1, RangeFloat1): RangeFloat1
}
TYPES_FLOORDIV: Dict[Tuple[Type, Type], Type] = {
    (int, int): int,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.timedelta, datetime.timedelta): int,
    (datetime.timedelta, int): int,
}
TYPES_TRUEDIV: Dict[Tuple[Type, Type], Type] = {
    (int, int): float,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (bool, int): int,
    (int, bool): int,
    (bool, float): float,
    (float, bool): float,
    (datetime.timedelta, datetime.timedelta): float,
    (datetime.timedelta, int): datetime.timedelta,
    (datetime.timedelta, float): datetime.timedelta,
}
TYPES_MOD = {
    (int, int): int,
    (float, int): float,
    (int, float): float,
    (float, float): float,
    (datetime.timedelta, datetime.timedelta): datetime.timedelta,
}
TYPES_ABS_NEG: Dict[Type, Type] = {
    int: int,
    float: float,
    datetime.timedelta: datetime.timedelta,
}
TYPES_CEIL_FLOOR_ROUND: Dict[Type, Type] = {
    int: int,
    float: int,
}
TYPES_LT_LE_GT_GE: Dict[Tuple[Type, Type], Type] = {
    (int, int): bool,
    (int, float): bool,
    (float, int): bool,
    (float, float): bool,
    (datetime.datetime, datetime.datetime): bool,
    (datetime.date, datetime.date): bool,
    (datetime.time, datetime.time): bool,
}



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import collections
import itertools
import json
import logging
from typing import List, Any, Generic, Type, Callable, Awaitable, Optional, Tuple, Union, Dict, Deque

from paho.mqtt.client import MQTTMessage, MQTTv311
from asyncio_mqtt import Client, MqttError, ProtocolVersion
from paho.mqtt.matcher import MQTTMatcher

from ._helper import SupervisedClientInterface
from ..base import Writable, Subscribable, T, S, LogicHandler
from ..conversion import SHCJsonEncoder, from_json

logger = logging.getLogger(__name__)


class MQTTClientInterface(SupervisedClientInterface):
    """
    SHC interface for connecting to MQTT brokers.

    The interface allows to create *subscribable* and *writable* connector objects for subscribing and publishing to
    individual MQTT topics. Additionally it provides generic MQTT client functionality, that can be used by other
    interfaces/adapters for implementing special protocols over MQTT (e.g. :class:`TasmotaInterface` for IoT devices
    running the Tasmota firmware).

    Internally, the interface uses a :class:`asyncio_mqtt.Client` object, which itself uses a
    :class:`paho.mqtt.client.Client`. The connection options (`hostname`, `port`, `username`, `password`, `client_id`
    and `protocol` are passed to that client implementation. TLS encrypted connections are currently not supported.

    By default, the interface features automatic reconnect on MQTT connection loss, but only if the initial connection
    attempt is successful. This behaviour can be customized using the `auto_reconnect` and `failsafe_start` parameters.

    For interacting with "raw" MQTT topics, use one of the following methods for creating connector objects:

      * :meth:`topic_raw` – for publishing and receiving raw :class:`bytes` payloads
      * :meth:`topic_string` – for publishing and receiving UTF-8-encoded strings
      * :meth:`topic_json` – for publishing and receiving arbitrary value types, using SHC's default JSON serialization
        (through :class:`shc.conversion.SHCJsonEncoder` resp. :func:`shc.conversion.from_json`)

    :param hostname: MQTT broker hostname or IP address
    :param port: MQTT broker port
    :param username: MQTT broker login username (no login is performed if None)
    :param password: MQTT broker login password (no login is performed if None)
    :param client_id: MQTT client name (if None, it is auto-generated by the Paho MQTT client)
    :param protocol: MQTT protocol version (from asyncio_mqtt library), defaults to MQTT 3.11
    :param auto_reconnect: If True (default), the client tries to reconnect automatically on connection errors with
        exponential backoff (1 * 1.25^n seconds sleep). Otherwise, the complete SHC system is shut down, when a
        connection error occurs.
    :param failsafe_start: If True and auto_reconnect is True, the API client allows SHC to start up, even if the MQTT
        connection can not be established in the first try. The connection is retried in background with exponential
        backoff (see `auto_reconnect` option). Otherwise (default), the first connection attempt on startup is not
        retried and will shutdown the SHC application on failure, even if `auto_reconnect` is True.
    """
    def __init__(self, hostname: str = 'localhost', port: int = 1883, username: Optional[str] = None,
                 password: Optional[str] = None, client_id: Optional[str] = None,
                 protocol: Union[ProtocolVersion, int] = ProtocolVersion.V311,
                 auto_reconnect: bool = True, failsafe_start: bool = False) -> None:
        super().__init__(auto_reconnect, failsafe_start)
        self.client = Client(hostname, port, username=username, password=password, client_id=client_id,
                             protocol=ProtocolVersion(protocol))
        self.matcher = MQTTMatcher()
        self.subscribe_topics: Dict[str, int] = {}  #: dict {topic: qos} for subscribing
        self.run_task: Optional[asyncio.Task] = None

    async def _connect(self) -> None:
        logger.info("Connecting MQTT client interface ...")
        # TODO this is a dirty hack to work around asyncio-mqtt's reconnect issues
        if self.client._disconnected.done():
            exc = self.client._disconnected.exception()
            if exc:
                logger.warning("MQTT client has been disconnected with exception:", exc_info=exc)
            self.client._connected = asyncio.Future()
            self.client._disconnected = asyncio.Future()
        await self.client.connect()

    async def _subscribe(self) -> None:
        if self.subscribe_topics:
            logger.info("Subscribing to MQTT topics ...")
            logger.debug("Topics: %s", self.subscribe_topics)
            await self.client.subscribe(list(self.subscribe_topics.items()))
        else:
            logger.info("No MQTT topics need to be subscribed")

    async def _disconnect(self) -> None:
        logger.info("Disconnecting MQTT client interface ...")
        try:
            await self.client.disconnect()
        except MqttError as e:
            if e.args[0] == 'Operation timed out':
                await self.client.force_disconnect()

    def topic_raw(self, topic: str, subscribe_topics: Optional[str] = None, qos: int = 0, retain: bool = False,
                  force_mqtt_subscription: bool = False) -> "RawMQTTTopicVariable":
        """
        Create a connector for publishing and receiving MQTT messages to/form a single topic as raw `bytes`

        :param topic: The MQTT topic to publish messages to
        :param subscribe_topics: The topic filter to subscribe to. If None (default) it will be equal to `topic`. This
            can be used to subscribe to a wildcard topic filter instead of only the single publishing topic.
            Attention: `subscribe_topics` must always include the `topic`!
        :param qos: The MQTT QoS for publishing messages to that topic (subscription QoS is currently always 0)
        :param retain: The MQTT *retain* bit for publishing messages to that topic
        :param force_mqtt_subscription: Force subscription to the MQTT topic(s) even if there are no SHC-internal
            *Subscribers* or logic handlers registered to this object. This is required for the `write` method to await
            the successful publishing of a published message to all clients. If there are local *Subscribers* or
            triggered logic handlers and `force_mqtt_subscription` is False, the client will not subscribe to the MQTT
            broker for this topic, so `write()` of this connector object will return immediately.
        """
        if subscribe_topics is None:
            subscribe_topics = topic
        return RawMQTTTopicVariable(self, topic, subscribe_topics, qos, retain, force_mqtt_subscription)

    def topic_string(self, topic: str, subscribe_topics: Optional[str] = None, qos: int = 0, retain: bool = False,
                     force_mqtt_subscription: bool = False) -> "StringMQTTTopicVariable":
        """
        Create a connector for publishing and receiving MQTT messages to/form a single topic as UTF-8 encoded strings

        :param topic: The MQTT topic to publish messages to
        :param subscribe_topics: The topic filter to subscribe to. If None (default) it will be equal to `topic`. This
            can be used to subscribe to a wildcard topic filter instead of only the single publishing topic.
            Attention: `subscribe_topics` must always include the `topic`!
        :param qos: The MQTT QoS for publishing messages to that topic (subscription QoS is currently always 0)
        :param retain: The MQTT *retain* bit for publishing messages to that topic
        :param force_mqtt_subscription: Force subscription to the MQTT topic(s) even if there are no SHC-internal
            *Subscribers* or logic handlers registered to this object. This is required for the `write` method to await
            the successful publishing of a published message to all clients. If there are local *Subscribers* or
            triggered logic handlers and `force_mqtt_subscription` is False, the client will not subscribe to the MQTT
            broker for this topic, so `write()` of this connector object will return immediately.
        """
        if subscribe_topics is None:
            subscribe_topics = topic
        return StringMQTTTopicVariable(self, topic, subscribe_topics, qos, retain, force_mqtt_subscription)

    def topic_json(self, type_: Type[T], topic: str, subscribe_topics: Optional[str] = None, qos: int = 0,
                   retain: bool = False, force_mqtt_subscription: bool = False) -> "JSONMQTTTopicVariable[T]":
        """
        Create a connector for publishing and receiving arbitrary values as JSON-encoded MQTT messages

        The connector uses SHC's generic JSON encoding/decoding system from the :mod:`shc.conversion` module. It has
        built-in support for basic datatypes, :class:`typing.NamedTuple`-based types and :class:`enum.Enum`-based types.
        For handling a custom JSON message format, you should create a custom Python type and specify JSON encoder and
        decoder functions for it, using :func:`shc.conversion.register_json_conversion`. Additionally, you'll want to
        add default type conversion functions from/to default types, using :func:`register_converter`, so that type
        conversion can happen on-the-fly between *connected* objects.

        :param type_: The connector's value type (used for its ``.type`` attribute and for chosing the correct JSON
            decoding function)
        :param topic: The MQTT topic to publish messages to
        :param subscribe_topics: The topic filter to subscribe to. If None (default) it will be equal to `topic`. This
            can be used to subscribe to a wildcard topic filter instead of only the single publishing topic.
            Attention: `subscribe_topics` must always include the `topic`!
        :param qos: The MQTT QoS for publishing messages to that topic (subscription QoS is currently always 0)
        :param retain: The MQTT *retain* bit for publishing messages to that topic
        :param force_mqtt_subscription: Force subscription to the MQTT topic(s) even if there are no SHC-internal
            *Subscribers* or logic handlers registered to this object. This is required for the `write` method to await
            the successful publishing of a published message to all clients. If there are local *Subscribers* or
            triggered logic handlers and `force_mqtt_subscription` is False, the client will not subscribe to the MQTT
            broker for this topic, so `write()` of this connector object will return immediately.
        """
        if subscribe_topics is None:
            subscribe_topics = topic
        return JSONMQTTTopicVariable(type_, self, topic, subscribe_topics, qos, retain, force_mqtt_subscription)

    async def publish_message(self, topic: str, payload: bytes, qos: int = 0, retain: bool = False) -> None:
        """
        Generic coroutine for publishing an MQTT message to be used by higher-level interfaces

        :param topic: The MQTT topic to send the message to
        :param payload: The raw MQTT message payload
        :param qos: The MQTT QoS value of the published message
        :param retain: The MQTT retain bit for publishing
        """
        logger.debug("Sending MQTT message t: %s p: %s q: %s r: %s", topic, payload, qos, retain)
        await self.client.publish(topic, payload, qos, retain)

    def register_filtered_receiver(self, topic_filter: str, receiver: Callable[[MQTTMessage], None],
                                   subscription_qos: int = 0) -> None:
        """
        Subscribe to an MQTT topic filter and register a callback function to be called when a message matching this
        filter is received.

        Warning: If multiple receivers with intersecting topic filters are registered, the receiver behaviour for MQTT
        messages seems to be undefined. Each receiver will still only receive the correct messages, but QoS guarantees
        will probably not be hold. Esp., messages might be received multiple times.

        :param topic_filter: An MQTT topic filter, possible containing wildcard characters. It is used for subscribing
            to these topics and for filtering the received messages.
        :param receiver: The callback function to be called when a matching MQTT message is received. The function
            must have a single free positional argument for taking the MQTTMessage object.
        :param subscription_qos: The subscription QoS
        """
        # If the topic is already subscribed (e.g. by another connector), merge the QOS values (in doubt, promote
        # it to 2)
        if topic_filter in self.subscribe_topics:
            self.subscribe_topics[topic_filter] = \
                2 if self.subscribe_topics[topic_filter] != subscription_qos else subscription_qos
        else:
            self.subscribe_topics[topic_filter] = subscription_qos
        try:
            receiver_list = self.matcher[topic_filter]
            receiver_list.append(receiver)
        except KeyError:
            self.matcher[topic_filter] = [receiver]

    async def _run(self) -> None:
        async with self.client.unfiltered_messages() as messages:
            self._running.set()
            async for message in messages:
                logger.debug("Incoming MQTT message: %s", message)
                receivers: List[List[Callable[[MQTTMessage], Awaitable[None]]]] \
                    = self.matcher.iter_match(message.topic)
                for receiver in itertools.chain.from_iterable(receivers):
                    receiver(message)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.client._hostname})"


class AbstractMQTTTopicVariable(Writable[T], Subscribable[T], Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, interface: MQTTClientInterface, publish_topic: str, subscribe_topics: str, qos: int,
                 retain: bool, force_mqtt_subscription: bool) -> None:
        super().__init__()
        self.interface = interface
        self.publish_topic = publish_topic
        self.subscribe_topics = subscribe_topics
        if not check_topic_matches_filter(publish_topic, subscribe_topics):
            raise ValueError("The publishing topic must match the subscribe topics filter! (Use an additional connector"
                             "otherwise.)")
        self.qos = qos
        self.retain = retain
        self._receiver_registered = False
        if force_mqtt_subscription:
            self.__subscribe_mqtt()
        self._pending_mqtt_pubs: Dict[bytes, Deque[asyncio.Event]] = {}

    def subscribe(self, subscriber: Writable[S], convert: Union[Callable[[T], S], bool] = False) -> None:
        self.__subscribe_mqtt()
        super().subscribe(subscriber, convert)

    def trigger(self, target: LogicHandler, synchronous: bool = False) -> LogicHandler:
        self.__subscribe_mqtt()
        return super().trigger(target, synchronous)

    def __subscribe_mqtt(self):
        if not self._receiver_registered:
            self.interface.register_filtered_receiver(self.subscribe_topics, self._new_value_from_mqtt)
            self._receiver_registered = True

    @abc.abstractmethod
    def _encode(self, value: T) -> bytes:
        pass

    @abc.abstractmethod
    def _decode(self, value: bytes) -> T:
        pass

    async def _write(self, value: T, origin: List[Any]) -> None:
        encoded_value = self._encode(value)
        if self._receiver_registered:
            queue = self._pending_mqtt_pubs.get(encoded_value)
            if queue is None:
                queue = collections.deque()
                self._pending_mqtt_pubs[encoded_value] = queue
            event = asyncio.Event()
            queue.append(event)

        await self.interface.publish_message(self.publish_topic, encoded_value, self.qos, self.retain)

        if self._receiver_registered:
            try:
                await asyncio.wait_for(event.wait(), 5)
            except asyncio.TimeoutError:
                logger.warning("Sent MQTT message was not received back from the broker within 5s.")
            finally:
                self._publish(value, origin)
                assert queue is not None
                queue.remove(event)
                if not queue:
                    del self._pending_mqtt_pubs[encoded_value]

    def _new_value_from_mqtt(self, message: MQTTMessage) -> None:
        if message.topic != self.publish_topic:
            self._publish(self._decode(message.payload), [])
            return

        queue = self._pending_mqtt_pubs.get(message.payload)
        if queue is not None:
            queue[0].set()
        else:
            self._publish(self._decode(message.payload), [])


class RawMQTTTopicVariable(AbstractMQTTTopicVariable[bytes]):
    type = bytes

    def _encode(self, value: bytes) -> bytes:
        return value

    def _decode(self, value: bytes) -> bytes:
        return value


class StringMQTTTopicVariable(AbstractMQTTTopicVariable[str]):
    type = str

    def _encode(self, value: str) -> bytes:
        return value.encode('utf-8')

    def _decode(self, value: bytes) -> str:
        return value.decode('utf-8-sig')


class JSONMQTTTopicVariable(AbstractMQTTTopicVariable[T], Generic[T]):
    def __init__(self, type_: Type[T], interface: MQTTClientInterface, publish_topic: str, subscribe_topics: str,
                 qos: int, retain: bool, force_mqtt_subscription: bool) -> None:
        self.type = type_
        super().__init__(interface, publish_topic, subscribe_topics, qos, retain, force_mqtt_subscription)

    def _encode(self, value: T) -> bytes:
        return json.dumps(value, cls=SHCJsonEncoder).encode('utf-8')

    def _decode(self, value: bytes) -> T:
        return from_json(self.type, json.loads(value.decode('utf-8-sig')))


def check_topic_matches_filter(topic: str, topic_filter: str) -> bool:
    """Returns true if the topic matches the topic_filter (which may contain wildcards).
    It can also be used to check if one topic_filter represents a subset of topics of another filter"""
    for a, b in itertools.zip_longest(topic.split('/'), topic_filter.split('/'), fillvalue=None):
        if b == '#':
            return True
        if b == '+' and a is not None:
            continue
        if a != b:
            return False
    return True



# Copyright 2020 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import datetime
import logging
from typing import Generic, Type, List, Any, Optional, Callable

from shc import conversion
from shc.base import Readable, Subscribable, Writable, handler, T, ConnectableWrapper, UninitializedError, Reading, S
from shc.datatypes import RangeFloat1, FadeStep
from shc.expressions import ExpressionWrapper
from shc.timer import Every

logger = logging.getLogger(__name__)

"""
This module contains some helper/adapter classes to support special patterns of interconnection of :class:`Connectable`
objects.
"""


class PeriodicReader(Readable[T], Subscribable[T], Generic[T]):
    """
    Wraps a :class:`Readable` object to turn it into :class:`Subscribable` object by periodically reading and publishing
    its value.

    :param wrapped: The *Readable* object, which shall be wrapped and read periodically
    :param interval: The interval in which the object's `.read()` coroutine is called and the result is published, e.g.
        ``datetime.timedelta(seconds=15)``.
    """
    def __init__(self, wrapped: Readable[T], interval: datetime.timedelta):
        self.type = wrapped.type
        super().__init__()
        self.wrapped = wrapped
        self.timer = Every(interval)
        self.timer.trigger(self.do_read, synchronous=True)

    async def read(self) -> T:
        return await self.wrapped.read()

    async def do_read(self, _value, origin) -> None:
        # We add the wrapped Readable object to the `origin` list to avoid publishing back its own value, in case it is
        # subscribed to one of our subscribers (e.g. a variable).
        self._publish(await self.wrapped.read(), origin + [self.wrapped])


class TwoWayPipe(ConnectableWrapper[T], Generic[T]):
    """
    A helper to connect two sets of Writable+Subscribable objects, without connecting the objects within each set. This
    object can be thought of as a bidirectional pipe: All write-events send to the left end of the pipe are forwarded to
    all objects on the right side and vice versa. This can be especially useful to connect a variable to multiple
    addresses of a home automation bus (like a KNX bus) without all mirroring incoming values on one address (e.g. a
    central or feedback datapoint) to the other addresses.

    To connect objects to one side of the pipe, use :meth:`connect_left` resp. :meth:`connect_right`. The additional
    method :meth:`connect` is an alias for `connect_left` to allow using a `TwoWayPipe` object as an argument for
    :meth:`Connectable.connect` itself with the result of connecting the pipe's left end to the object.

    The following example demonstrates, how to connect two interface connectors to a Variable, such that the Variable
    will interact with both of them, without forwarding events/values from one connector to the other::

        shc.Variable(bool)\\
            .connect(TwoWayPipe(bool)
                .connect_right(some_interface.connector(1))
                .connect_right(some_interface.connector(2)))

    :param type_: The `type` of the values to be forwarded. This is used as the `type` of the two pipe-end *connectable*
        objects.
    """
    def __init__(self, type_: Type[T]):
        self.type = type_
        self.left = _PipeEnd(type_)
        self.right = _PipeEnd(type_)
        self.left.other_end = self.right
        self.right.other_end = self.left

    def connect_right(self, *args, **kwargs) -> "TwoWayPipe":
        self.right.connect(*args, **kwargs)
        return self

    def connect_left(self, *args, **kwargs) -> "TwoWayPipe":
        self.left.connect(*args, **kwargs)
        return self

    def connect(self, *args, **kwargs) -> "TwoWayPipe":
        self.left.connect(*args, **kwargs)
        return self


class _PipeEnd(Subscribable[T], Writable[T], Generic[T]):
    def __init__(self, type_: Type[T]):
        self.type = type_
        super().__init__()
        self.other_end: "_PipeEnd" = None  # type: ignore

    async def _write(self, value: T, origin: List[Any]):
        await self.other_end._publish_and_wait(value, origin)


class BreakableSubscription(Subscribable[T], Generic[T]):
    """
    A transparent wrapper for `Subscribable` objects, that allows to dynamically switch forwarding of new values on and
    off.

    A `BreakableSubscription` object can wrap any `Subscribable` object and is `Subscribable` itself to transparently
    re-publish all published values of the wrapped object. However, the re-publishing can be switched on and off through
    a `Readable` boolean control object.

    This can be used to dynamically disable automatic control rules, e.g. control of variables through expressions.
    In the following example, the `fan` is automatically controlled via an expression, based on `room_temperature`, but
    only as long as the `automatic_fan_control` is on. Otherwise, it will not get value updates from that comparison
    expression and thus keep its current value::

        fan = shc.Variable(bool)
        room_temperature = shc.Variable(float)
        automatic_fan_control = shc.Variable(bool)

        fan.connect(shc.misc.BreakableSubscription(room_temperature.EX > 25.0,
                                                   automatic_fan_control))

    If the wrapped object is also `Readable` and the control object is also `Subscribable`, the current value of the
    wrapped object is read and published when the subscription is enabled (the control object changes to `True`).

    If you want to select a value from two or more `Readable` and `Subscribable` objects dynamically, take a look at
    :class:`shc.expressions.IfThenElse` or :class:`shc.expressions.Multiplexer`.

    :param wrapped: The Subscribable object to be wrapped
    :param control: The Readable control object
    """
    _synchronous_publishing = True

    def __init__(self, wrapped: Subscribable[T], control: Readable[bool]):
        self.type = wrapped.type
        super().__init__()
        self.wrapped = wrapped
        self.control = control
        wrapped.trigger(self._new_value, synchronous=True)
        if isinstance(control, Subscribable) and isinstance(wrapped, Readable):
            control.trigger(self._connection_change, synchronous=True)

    async def _new_value(self, value: T, origin: List[Any]) -> None:
        try:
            connected = await self.control.read()
        except UninitializedError:
            # We default to False, if the control variable is not initialized yet
            return
        if connected:
            await self._publish_and_wait(value, origin)

    async def _connection_change(self, connected: bool, origin: List[Any]) -> None:
        if not connected:
            return
        try:
            await self._publish_and_wait(await self.wrapped.read(), origin)  # type: ignore
        except UninitializedError:
            pass


class Hysteresis(Subscribable[bool], Readable[bool]):
    """
    A Hysteresis function wrapper for Subscribable objects of any comparable type.

    The `Hysteresis` object is a `Subscribable` and `Readable` object of boolean type, which wraps any `Subscribable`
    object of a `type` which is comparable with `<`. On any value update of the wrapped object, its current value is
    compared to two fixed bounds. When the value exceeds the upper bound, the `Hysteresis` publishes a `True` value;
    when it falls below the lower bound, the `Hysteresis` publishes a `False` value. As long as the value stays between
    the two bounds, the `Hysteresis` keeps its previous value (and does not publish anything). The output boolean
    value of the `Hysteresis` may be inverted via the `inverted` parameter.

    :param wrapped: The `Subscribable` object to apply the hysteresis on
    :param lower: The lower bound of the hysteresis
    :param upper: The lower bound of the hysteresis
    :param inverted: If True, the boolean output is inverted (False when value is above upper bound, True when value is
        below lower bound)
    :param initial_value: The initial value of the `Hysteresis` object, which is returned on `read()` requests, until
        the first value outside of the bounds is received. Attention: If `inverted` is True, the initial value is
        inverted, too.
    """
    type = bool

    def __init__(self, wrapped: Subscribable[T], lower: T, upper: T, inverted: bool = False,
                 initial_value: bool = False):
        super().__init__()
        wrapped.trigger(self._new_value, synchronous=True)
        self._value = initial_value  #: Current output value (uninverted)
        if not isinstance(lower, wrapped.type) or not isinstance(upper, wrapped.type):
            raise TypeError("'lower' and 'upper' must be instances of the wrapped Subscribable's type, which is {}"
                            .format(wrapped.type.__name__))
        self.lower = lower
        self.upper = upper
        if lower > upper:  # type: ignore  # To defined that T is comparable, we need to use Protocols from Python 3.8
            raise ValueError('Lower bound of hysteresis must be lower than upper bound.')
        self.inverted = inverted

    async def _new_value(self, value: T, origin: List[Any]) -> None:
        old_value = self._value
        if value < self.lower:  # type: ignore
            self._value = False
        elif value > self.upper:  # type: ignore
            self._value = True
        if self._value != old_value:
            await self._publish_and_wait(self._value != self.inverted, origin)

    async def read(self) -> bool:
        return self._value != self.inverted

    @property
    def EX(self) -> ExpressionWrapper[bool]:
        return ExpressionWrapper(self)


class FadeStepAdapter(Subscribable[RangeFloat1], Reading[RangeFloat1]):
    """
    An adapter to connect a Subscribable object which publishes :class:`shc.datatypes.FadeStep` values to a
    :class:`shc.datatypes.RangeFloat1` Variable (or another Readable + Wrtiable object of type `RangeFloat1`), such that
    the FadeSteps are applied to the current value of that object.

    Example usage::

        dim_up_button = shc.web.widgets.StatelessButton(FadeStep(0.1), "up")
        dim_down_button = shc.web.widgets.StatelessButton(FadeStep(-0.1), "down")

        dimmer_value = shc.Variable(RangeFloat1)\
            .connect(FadeStepAdapter(dim_up_button))\
            .connect(FadeStepAdapter(dim_down_button))

    To apply FadeSteps as dynamic ramps of a fixed duration instead of a sudden jump, take a look at
    :class:`shc.timer.FadeStepRamp` (see description in the documentation of :class:`shc.timer.AbstractRamp`).

    :param wrapped: The Subscribable object which shall be wrapped to apply its published FadeStep values to connected
        objects
    """
    type = RangeFloat1
    is_reading_optional = False

    def __init__(self, wrapped: Subscribable[FadeStep]):
        super().__init__()
        if not issubclass(wrapped.type, FadeStep):
            raise TypeError("First Parameter to `FadeStepAdapter` must be a Subscribable object with value type "
                            "`FadeStep`")
        wrapped.trigger(self._update, synchronous=True)

    async def _update(self, value: FadeStep, origin: List[Any]) -> None:
        current_value = await self._from_provider()
        if current_value is None:
            logger.warning("Cannot apply FadeStep, since current value is not available.")
            return
        await self._publish_and_wait(value.apply_to(current_value), origin)


class ConvertSubscription(Subscribable[T], Generic[T]):
    """
    An adapter, wrapping a subscribable object of value type S and converting it to a subscribable object of value type
    T, using the default converter or a given converter function.

    Typically, the following methods should be used instead of this class:

    - Use :meth:`.subscribe() <shc.base.Subscribable.subscribe>`'s ``convert`` parameter
    - Use an :ref:`SHC Expression <expressions>` and its :meth:`.convert() <shc.expressions.ExpressionBuilder.convert>`
      method

    However, they may be situations, where neither of these methods is applicable, e.g. when the subscribable object is
    not readable (so it should not used in an expression) and it is not used with `.subscribe()` or `.connect()`
    directly (e.g. it is passed to another object's constructor).
    """

    def __init__(self, wrapped: Subscribable[S], type_: Type[T], convert: Optional[Callable[[S], T]] = None):
        self.type = type_
        super().__init__()
        if convert is None:
            self.converter = conversion.get_converter(wrapped.type, type_)
        else:
            self.converter = convert
        wrapped.trigger(self.__update, synchronous=True)

    async def __update(self, value: S, origin: List[Any]):
        await self._publish_and_wait(self.converter(value), origin)



"""Command interface."""
import asyncio
import logging
from typing import List, Union, cast

from shc.base import Readable, UninitializedError

logger = logging.getLogger(__name__)


class Command(Readable[str]):
    """
    A *Readable* object that executes a (fixed) given command returns the result when being read.

    This *readable* object can easily be combined with the :class:`shc.misc.PeriodicReader` to periodically publish the
    command's output::

        # Publish "Hello, world!" every 5seconds (the hard way ;) )
        PeriodicReader(Command(['echo', 'Hello, world!']), datetime.timedelta(seconds=5))

    The parameters are similar to subprocess.run()'s:

    :param command: The command to execute incl. arguments and parameters. A single str, when `shell` is True, otherwise
        it should be list of the command and its individual command line arguments.
    :param shell: If True, the command will be within a shell, otherwise it will be started as a plain subprocess
    :param include_std_err: If True, stderr output from the command will be included in the output. This is similar to
        adding '2>&1' to your shell command
    :param check: If True, the exit code of the command will be checked the read() method raises an UninitializedError
        instead of returning the output of the command in case of a non-zero exit code.
    """
    type = str

    def __init__(self, command: Union[str, List[str]], shell=False, include_std_err=False, check=False):
        """Initalize the command interface."""
        super().__init__()
        self.command = command
        self.shell = shell
        assert isinstance(self.command, str) or not self.shell
        self.include_std_err = include_std_err
        self.check = check

    async def read(self) -> str:
        stderr = asyncio.subprocess.STDOUT if self.include_std_err else asyncio.subprocess.DEVNULL
        if self.shell:
            command = cast(str, self.command)
            command_process = await asyncio.create_subprocess_shell(command,
                                                                    stdout=asyncio.subprocess.PIPE,
                                                                    stderr=stderr)
        else:
            command_args = cast(List[str], self.command)
            command_process = await asyncio.create_subprocess_exec(*command_args,
                                                                   stdout=asyncio.subprocess.PIPE,
                                                                   stderr=stderr)

        std_out, _std_err = await command_process.communicate()
        if self.check:
            if command_process.returncode != 0:
                logger.warning("Subprocess %s returned non-zero exit code %s", self.command, command_process.returncode)
                raise UninitializedError()

        std_out_str = std_out.decode().strip()
        logger.debug("Command output: %s", std_out)
        return std_out_str


class CommandExitCode(Readable[int]):
    """
    A *Readable* object that executes a (fixed) given command returns the its exit code when being read.

    This *readable* object can easily be combined with the :class:`shc.misc.PeriodicReader` to periodically publish the
    command's exit code::

        # Publish 1 every 5seconds (the hard way ;) )
        PeriodicReader(Command(['false']), datetime.timedelta(seconds=5))

    The parameters are similar to subprocess.run()'s:

    :param command: The command to execute incl. arguments and parameters. A single str, when `shell` is True, otherwise
        it should be list of the command and its individual command line arguments.
    :param shell: If True, the command will be within a shell, otherwise it will be started as a plain subprocess
    """
    type = int

    def __init__(self, command: Union[str, List[str]], shell=False):
        """Initalize the command interface."""
        super().__init__()
        self.command = command
        self.shell = shell

    async def read(self) -> int:
        if self.shell:
            command = cast(str, self.command)
            command_process = await asyncio.create_subprocess_shell(command,
                                                                    stdout=asyncio.subprocess.DEVNULL,
                                                                    stderr=asyncio.subprocess.DEVNULL)
        else:
            command_args = cast(List[str], self.command)
            command_process = await asyncio.create_subprocess_exec(*command_args,
                                                                   stdout=asyncio.subprocess.DEVNULL,
                                                                   stderr=asyncio.subprocess.DEVNULL)

        result = await command_process.wait()
        return result



# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import collections
import functools
import json
import logging
import time
from typing import List, Any, Dict, Deque, Generic, Union, Type, TypeVar, Tuple, cast, Optional

from paho.mqtt.client import MQTTMessage

from ..base import Writable, Subscribable, T, Readable
from .mqtt import MQTTClientInterface
from ..datatypes import RangeInt0To100, RGBUInt8, RangeUInt8, RGBWUInt8, RGBCCTUInt8, CCTUInt8
from ..supervisor import AbstractInterface, InterfaceStatus, ServiceStatus

logger = logging.getLogger(__name__)

ConnType = TypeVar("ConnType", bound="AbstractTasmotaConnector")
JSONType = Union[str, float, int, None, Dict[str, Any], List[Any]]


class TasmotaInterface(AbstractInterface):
    """
    SHC interface to connect with ESP8266-based IoT devices, running the
    `Tasmota firmware <https://tasmota.github.io/>`_, via MQTT.

    Requires a :class:`MQTTClientInterface` which is connected to the MQTT broker to which your Tasmota device(s)
    connect. Each instance of `TasmotaInterface` connects with a single Tasmota device. It identifies the individual
    Tasmota device via its MQTT "Topic" (by default something like "tasmota_A0B1C2").

    The `TasmotaInterface` provides *subscribable* and (mostly) *writable* connector objects to receive state updates
    from the Tasmota device and send commands. Currently, only PWM-dimmed lights and IR receivers are fully supported.
    Please file a GitHub issue if you are interested in other Tasmota functionality.

    :param mqtt_interface: The `MQTTClientInterface` to use for MQTT communication
    :param device_topic: The Tasmota devices individual "Topic", used to address the device, as it is configured on the
        Tasmota web UI at Configuration → MQTT or using Tasmota's `Topic` command. By default it should look like
        "tasmota_A0B1C2".
    :param topic_template: The Tasmota "Full Topic", used build MQTT topics from the individual device's topic.
        Only required if you don't use the default Full Topic "%prefix%/%topic%/". In contrast to the Tasmota web UI
        or "FullTopic" command, the parameter uses Python's modern formatting syntax: If your Tasmota "Full Topic" is
        "devices/%topic%/%prefix%/", you must specify "devices/{topic}/{prefix}/" as `topic_tempalate`.
    :param telemetry_interval: The expected interval of periodic telemetry messages from the Tasmota device in seconds.
        This is used by the status monitoring to detect device failures. Use `0` to disable telemetry monitoring.
    """
    def __init__(self, mqtt_interface: MQTTClientInterface, device_topic: str,
                 topic_template: str = "{prefix}/{topic}/", telemetry_interval: int = 300):
        super().__init__()
        self.mqtt_interface = mqtt_interface
        self.device_topic = device_topic
        self.topic_template = topic_template
        self.telemetry_interval = telemetry_interval
        self._connectors_by_result_field: Dict[str, List[AbstractTasmotaConnector]] = {}
        self._connectors_by_type: Dict[Type[AbstractTasmotaConnector], AbstractTasmotaConnector] = {}
        self._pending_commands: Deque[Tuple[str, List[Any], asyncio.Event]] = collections.deque()
        self._online_connector = TasmotaOnlineConnector()

        self._latest_telemetry_time: Optional[float] = None
        self._latest_telemetry: Optional[Dict[str, JSONType]] = None

        # Subscribe relevant MQTT topics and register message handlers
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='tele', topic=device_topic) + 'RESULT',
                                                  self._handle_result_or_status, 1)
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='stat', topic=device_topic) + 'RESULT',
                                                  functools.partial(self._handle_result_or_status, result=True), 1)
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='tele', topic=device_topic) + 'STATE',
                                                  self._handle_result_or_status, 1)
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='tele', topic=device_topic) + 'SENSOR',
                                                  self._handle_result_or_status, 1)
        # TODO allow listening to STATUS8 responses for sensors
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='stat', topic=device_topic) + 'STATUS11',
                                                  self._handle_status11, 1)
        mqtt_interface.register_filtered_receiver(topic_template.format(prefix='tele', topic=device_topic) + 'LWT',
                                                  self._online_connector._update_from_mqtt, 1)

    async def start(self) -> None:
        # Send status request (for telemetry data and state) as soon as the MQTT interface is up
        await self.mqtt_interface.wait_running(5)
        await self._send_command("status", "11")
        # TODO send STATUS8 cmnd?

    async def stop(self) -> None:
        pass

    async def get_status(self) -> InterfaceStatus:
        # Check Tasmota online state (via Last Will message)
        if not self._online_connector.value:
            return InterfaceStatus(ServiceStatus.CRITICAL, "Tasmota device is not online")

        # Check telemetry data
        if not self._latest_telemetry:
            if not self.telemetry_interval:
                return InterfaceStatus()
            return InterfaceStatus(ServiceStatus.CRITICAL, "No telemetry data received from Tasmota device by now")
        assert self._latest_telemetry_time
        last_telemetry_age = time.monotonic() - self._latest_telemetry_time
        if last_telemetry_age > 10 * self.telemetry_interval:
            return InterfaceStatus(ServiceStatus.CRITICAL, "No telemetry data received from Tasmota device by now")
        indicators = {
            'telemetry_age': last_telemetry_age,
            'tasmota.UptimeSec': self._latest_telemetry.get('UptimeSec'),
            'tasmota.Heap': self._latest_telemetry.get('Heap'),
            'tasmota.LoadAvg': self._latest_telemetry.get('LoadAvg'),
            'tasmota.Wifi.RSSI': self._latest_telemetry.get('Wifi', {}).get('RSSI'),  # type: ignore
            'tasmota.Wifi.Signal': self._latest_telemetry.get('Wifi', {}).get('Signal'),   # type: ignore
            'tasmota.Wifi.Downtime': self._latest_telemetry.get('Wifi', {}).get('Downtime'),  # type: ignore
        }
        if last_telemetry_age > 10 * self.telemetry_interval:
            return InterfaceStatus(status=ServiceStatus.CRITICAL,
                                   message="Latest telemetry data from Tasmota device is {:.2f}s old (more than 10x the"
                                           " expected telemetry interval)".format(last_telemetry_age),
                                   indicators=indicators)  # type: ignore
        if last_telemetry_age > 1.5 * self.telemetry_interval:
            return InterfaceStatus(status=ServiceStatus.WARNING,
                                   message="Latest telemetry data from Tasmota device is {:.2f}s old (more than 1.5x "
                                           "the expected telemetry interval)".format(last_telemetry_age),
                                   indicators=indicators)  # type: ignore
        return InterfaceStatus(indicators=indicators)  # type: ignore

    def _handle_result_or_status(self, msg: MQTTMessage, result: bool = False) -> None:
        """
        Callback function to handle incoming MQTTMessages on the Tasmota device's RESULT, STATUS and STATE topics

        :param msg: The MQTTMessage to be parsed and handled.
        :param result: Shall be True if the handled message has been published on the Tasmota RESULT topic (i.e. it is
            probably a result to a Tasmota command we issued recently)
        """
        try:
            data = json.loads(msg.payload.decode('utf-8'))
            assert isinstance(data, dict)
        except (json.JSONDecodeError, UnicodeDecodeError, AssertionError) as e:
            logger.error("Could not decode Tasmota result as JSON object: %s", msg.payload, exc_info=e)
            return
        self._dispatch_status(data, result)

    def _handle_status11(self, msg: MQTTMessage) -> None:
        """
        Callback function to handle incoming MQTTMessages on the STATUS11 topic (as a result to of the 'status 11')
        command.
        """
        try:
            data = json.loads(msg.payload.decode('utf-8'))
            assert isinstance(data, dict)
            assert 'StatusSTS' in data
            assert isinstance(data['StatusSTS'], dict)
        except (json.JSONDecodeError, UnicodeDecodeError, AssertionError) as e:
            logger.error("Could not decode Tasmota telemetry status as JSON object: %s", msg.payload, exc_info=e)
            return
        self._dispatch_status(data['StatusSTS'], False)

    def _dispatch_status(self, data: Dict[str, JSONType], result: bool) -> None:
        """
        Internal helper method to dispatch results/telemetry updates and sensor readings from Tasmota device, received
        via :meth:`_handle_result_or_status` or :meth:`_handle_status11` for publishing by all affected connectors.

        :param data: The parsed JSON payload of the tasmota result/telemetry message
        :param result: Shall be True if the handled message has been published on the Tasmota RESULT topic (i.e. it is
            probably a result to a Tasmota command we issued recently)
        """
        logger.debug("Dispatching Tasmota result/status from %s: %s", self.device_topic, data)

        origin = []
        event = None

        # If the message has been received as a Tasmota "result", check if is the response to a recent command, in order
        # to publish the value updates with the correct 'origin' and let the sending Connector's `write()` method return
        # (by set()ing the associated Event).
        #
        # Notice, that we only check the field names to match received results with pending commands. I.e., we assume
        # that the first received result which contains the Tasmota field affected by an given command (and not matching
        # a previous command) belongs to that command. This may be wrong when a concurrent update of the same field (or
        # an internally linked field in Tasmota) from somewhere else happens. However, the worst outcome in that case
        # should be the unintended re-publishing of the result to our command to its origin within SHC. This should not
        # be an issue, since it will correctly represent the latest state of the Tasmota device and only happen once
        # (and thus not result in a feedback loop).
        if result:
            for field, origin_, event_ in self._pending_commands:
                if field in data:
                    origin = origin_
                    event = event_
                    logger.debug("The result/status is considered a result to our recent command for '%s' field, "
                                 "originating from %s", field, origin_)
                    break

        for key, value in data.items():
            for connector in self._connectors_by_result_field.get(key, []):
                try:
                    connector._publish(connector._decode(value), origin)
                except Exception as e:
                    logger.error("Error while processing Tasmota result/status field %s=%s from %s in Tasmota "
                                 "connector %s", key, value, self.device_topic, connector, exc_info=e)
        if event:
            event.set()

        # If it seems to be (periodic) telemetry update, store it for usage by our `get_status()` method
        if not result and "Uptime" in data:
            self._latest_telemetry = data
            self._latest_telemetry_time = time.monotonic()

    def online(self) -> "TasmotaOnlineConnector":
        """
        Returns a *readable* and *subscribable* :class:`bool`-typed Connector that will indicate the online-state of the
        Tasmota device, using its MQTT "Last Will Topic".
        """
        return self._online_connector

    def power(self) -> "TasmotaPowerConnector":
        """
        Returns a *subscribable* and *writable* :class:`bool`-typed Connector to monitor and control the POWER state of
        the Tasmota device.
        """
        return self._get_or_create_connector(TasmotaPowerConnector)

    def dimmer(self) -> "TasmotaDimmerConnector":
        """
        Returns a *subscribable* and *writable* Connector to monitor and control the Dimmer state of the Tasmota device.

        The dimmer value is an integer between 0 and 100, represented as :class:`shc.datatypes.RangeInt0To100` in SHC.
        """
        return self._get_or_create_connector(TasmotaDimmerConnector)

    def color_cct(self) -> "TasmotaColorCCTConnector":
        """
        Returns a *subscribable* and *writable* Connector to monitor and control two channel (cold-white, white-white)
        dimmable lamps attached to the Tasmota device. The value is represented as :class:`shc.datatypes.CCTUInt8`
        (composed of two 8-bit :class:`shc.datatypes.RangeUInt8` integer values).

        Note, that this connector is only sensible if the Tasmota device has two PWM pins for cold white and warm white
        configured. It will also work with different numbers of PWM channels configured, but the channel mapping will
        not be sensible.

        See https://tasmota.github.io/docs/Lights/ for more information on Tasmota light controls.
        """
        return self._get_or_create_connector(TasmotaColorCCTConnector)

    def color_rgb(self) -> "TasmotaColorRGBConnector":
        """
        Returns a *subscribable* and *writable* Connector to monitor and control three channel (red, green, blue)
        dimmable lamps attached to the Tasmota device. The value is represented as :class:`shc.datatypes.RGBUInt8`
        (composed of three 8-bit :class:`shc.datatypes.RangeUInt8` integer values).

        Note, that this connector is only sensible if the Tasmota device has at least three PWM pins for red, green and
        blue configured. It will also work with less PWM channels configured, but the channel mapping will not be
        sensible.

        See https://tasmota.github.io/docs/Lights/ for more information on Tasmota light controls.
        """
        return self._get_or_create_connector(TasmotaColorRGBConnector)

    def color_rgbw(self) -> "TasmotaColorRGBWConnector":
        """
        Returns a *subscribable* and *writable* Connector to monitor and control four channel (red, green, blue, white)
        dimmable lamps attached to the Tasmota device. The value is represented as :class:`shc.datatypes.RGBWUInt8`
        (composed of three 8-bit :class:`shc.datatypes.RangeUInt8` integer values for RGB in an
        `shc.datatypes.RGBUInt8` and an additional :class:`shc.datatypes.RangeUInt8` value for the white channel).

        Note, that this connector is only sensible if the Tasmota device has four PWM pins for red, green, blue and
        white configured. It will also work with a different number PWM channels configured, but the channel mapping
        will not always be sensible.

        See https://tasmota.github.io/docs/Lights/ for more information on Tasmota light controls.
        """
        return self._get_or_create_connector(TasmotaColorRGBWConnector)

    def color_rgbcct(self) -> "TasmotaColorRGBCCTConnector":
        """
        Returns a *subscribable* and *writable* Connector to monitor and control five channel (red, green, blue, cold
        white, warm white) dimmable lamps attached to the Tasmota device. The value is represented as
        :class:`shc.datatypes.RGBCCTUInt8`
        (composed of three 8-bit :class:`shc.datatypes.RangeUInt8` integer values for RGB in an
        `shc.datatypes.RGBUInt8` and two additional :class:`shc.datatypes.RangeUInt8` values in an
        class:`shc.datatypes.CCTUInt8`).

        Note, that this connector is only sensible if the Tasmota device has five PWM pins configured. It will also work
        less PWM channels configured, but the channel mapping will not always be sensible.

        See https://tasmota.github.io/docs/Lights/ for more information on Tasmota light controls.
        """
        return self._get_or_create_connector(TasmotaColorRGBCCTConnector)

    def ir_receiver(self) -> "TasmotaIRReceiverConnector":
        """
        Returns a *subscribable* :class:`bytes`-typed Connector that publishes the data/hash of each IR command received
        by the Tasmota device (via an IR receiver, attatched to a pin configured as `IRrecv`).

        See https://tasmota.github.io/docs/Tasmota-IR/#receiving-ir-commands for more information about receiving IR
        commands.
        """
        return self._get_or_create_connector(TasmotaIRReceiverConnector)

    def energy_power(self) -> "TasmotaEnergyPowerConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured power consumption in
        watts from the Tasmota device, if available.
        """
        return self._get_or_create_connector(TasmotaEnergyPowerConnector)

    def energy_voltage(self) -> "TasmotaEnergyVoltageConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured mains voltage in
        volts from the Tasmota device, if available.
        """
        return self._get_or_create_connector(TasmotaEnergyVoltageConnector)

    def energy_current(self) -> "TasmotaEnergyCurrentConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured current flow in
        amperes from the Tasmota device, if available.
        """
        return self._get_or_create_connector(TasmotaEnergyCurrentConnector)

    def energy_total(self) -> "TasmotaEnergyTotalConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the total energy consumption measured by the
        Tasmota device since its last reboot in kWh.
        """
        return self._get_or_create_connector(TasmotaEnergyTotalConnector)

    def energy_power_factor(self) -> "TasmotaEnergyFactorConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured power factor from the
        Tasmota device.
        """
        return self._get_or_create_connector(TasmotaEnergyFactorConnector)

    def energy_apparent_power(self) -> "TasmotaEnergyApparentPowerConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured apparent power in VA
        from the Tasmota device.
        """
        return self._get_or_create_connector(TasmotaEnergyApparentPowerConnector)

    def energy_reactive_power(self) -> "TasmotaEnergyReactivePowerConnector":
        """
        Returns a *subscribable* :class:`float`-typed Connector, publishing the currently measured reactive power in VAr
        from the Tasmota device.
        """
        return self._get_or_create_connector(TasmotaEnergyReactivePowerConnector)

    def _get_or_create_connector(self, type_: Type[ConnType]) -> ConnType:
        """
        Helper method to create or get the connector object of a given type, while making sure that there is only one
        of each type in each TasmotaInterface.

        :param type_: The connector type (class) to be created
        :return: The existing (or newly created) connector of the specified type
        """
        if type_ in self._connectors_by_type:
            return cast(ConnType, self._connectors_by_type[type_])
        else:
            conn = type_(self)  # type: ignore
            self._connectors_by_type[type_] = conn
            if conn.result_field in self._connectors_by_result_field:
                self._connectors_by_result_field[conn.result_field].append(conn)
            else:
                self._connectors_by_result_field[conn.result_field] = [conn]
            return conn

    async def _send_command(self, command: str, value: str):
        """
        Internal helper method to send a Tasmota command to the device via MQTT.

        See https://tasmota.github.io/docs/Commands/ for a reference of available commands.

        :param command: The Tasmota command (used as part of the MQTT topic)
        :param value: The parameters of the Tasmota command (used as MQTT payload)
        """
        # TODO raise exception if device is disconnected
        await self.mqtt_interface.publish_message(
            self.topic_template.format(prefix='cmnd', topic=self.device_topic) + command,
            value.encode())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.device_topic})"


class AbstractTasmotaConnector(Subscribable[T], Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, result_field: str):
        super().__init__()
        self.result_field = result_field

    @abc.abstractmethod
    def _decode(self, value: JSONType) -> T:
        pass


class AbstractTasmotaRWConnector(AbstractTasmotaConnector[T], Writable[T], Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, interface: TasmotaInterface, command: str, result_field: str):
        super().__init__(result_field)
        self.command = command
        self.interface = interface

    async def _write(self, value: T, origin: List[Any]) -> None:
        encoded_value = self._encode(value)
        event = asyncio.Event()

        self.interface._pending_commands.append((self.result_field, origin, event))

        try:
            await self.interface._send_command(self.command, encoded_value)
            await asyncio.wait_for(event.wait(), 5)
        except asyncio.TimeoutError:
            logger.warning("No Result from Tasmota device %s to %s command within 5s.", self.interface.device_topic,
                           self.command)
        finally:
            # Remove queue entry
            index = -1
            for pos, t in enumerate(self.interface._pending_commands):
                if t[2] is event:
                    index = pos
                    break
            assert index != -1
            del self.interface._pending_commands[index]

    @abc.abstractmethod
    def _encode(self, value: T) -> str:
        pass


class TasmotaPowerConnector(AbstractTasmotaRWConnector[bool]):
    type = bool

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "Power", "POWER")

    def _encode(self, value: bool) -> str:
        return 'ON' if value else 'OFF'

    def _decode(self, value: JSONType) -> bool:
        assert isinstance(value, str)
        return value.lower() in ('on', '1', 'true')


class TasmotaDimmerConnector(AbstractTasmotaRWConnector[RangeInt0To100]):
    type = RangeInt0To100

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "Dimmer", "Dimmer")

    def _encode(self, value: RangeInt0To100) -> str:
        return str(value)

    def _decode(self, value: JSONType) -> RangeInt0To100:
        assert isinstance(value, int)
        return RangeInt0To100(value)


class TasmotaColorCCTConnector(AbstractTasmotaRWConnector[CCTUInt8]):
    type = CCTUInt8

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "color", "Color")

    def _encode(self, value: CCTUInt8) -> str:
        return '#{:0>2X}{:0>2X}'.format(value.cold, value.warm)

    def _decode(self, value: JSONType) -> CCTUInt8:
        assert isinstance(value, str)
        data = bytes.fromhex(value)
        data += bytes([0] * (2 - len(data)))
        return CCTUInt8(RangeUInt8(data[0]), RangeUInt8(data[1]))


class TasmotaColorRGBConnector(AbstractTasmotaRWConnector[RGBUInt8]):
    type = RGBUInt8

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "color", "Color")

    def _encode(self, value: RGBUInt8) -> str:
        return '#{:0>2X}{:0>2X}{:0>2X}'.format(value.red, value.green, value.blue)

    def _decode(self, value: JSONType) -> RGBUInt8:
        assert isinstance(value, str)
        data = bytes.fromhex(value)
        data += bytes([0] * (3 - len(data)))
        return RGBUInt8(*(RangeUInt8(v) for v in data))


class TasmotaColorRGBWConnector(AbstractTasmotaRWConnector[RGBWUInt8]):
    type = RGBWUInt8

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "color", "Color")

    def _encode(self, value: RGBWUInt8) -> str:
        return '#{:0>2X}{:0>2X}{:0>2X}{:0>2X}'.format(value.rgb.red, value.rgb.green, value.rgb.blue, value.white)

    def _decode(self, value: JSONType) -> RGBWUInt8:
        assert isinstance(value, str)
        data = bytes.fromhex(value)
        data += bytes([0] * (4 - len(data)))
        return RGBWUInt8(RGBUInt8(*(RangeUInt8(v) for v in data[0:3])), RangeUInt8(data[3]))


class TasmotaColorRGBCCTConnector(AbstractTasmotaRWConnector[RGBCCTUInt8]):
    type = RGBCCTUInt8

    def __init__(self, interface: TasmotaInterface):
        super().__init__(interface, "color", "Color")

    def _encode(self, value: RGBCCTUInt8) -> str:
        return '#{:0>2X}{:0>2X}{:0>2X}{:0>2X}{:0>2X}'.format(value.rgb.red, value.rgb.green, value.rgb.blue,
                                                             value.white.cold, value.white.warm)

    def _decode(self, value: JSONType) -> RGBCCTUInt8:
        assert isinstance(value, str)
        data = bytes.fromhex(value)
        data += bytes([0] * (5 - len(data)))
        return RGBCCTUInt8(RGBUInt8(*(RangeUInt8(v) for v in data[0:3])),
                           CCTUInt8(RangeUInt8(data[3]), RangeUInt8(data[4])))


class TasmotaIRReceiverConnector(AbstractTasmotaConnector[bytes]):
    type = bytes

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("IrReceived")

    def _decode(self, value: JSONType) -> bytes:
        assert isinstance(value, dict)
        # TODO this is hacky. We should probably add the detected protocol
        data = value['Data'] if 'Data' in value else value['Hash']
        return bytes.fromhex(data[2:])


class TasmotaEnergyPowerConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['Power'])


class TasmotaEnergyVoltageConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['Voltage'])


class TasmotaEnergyCurrentConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['Current'])


class TasmotaEnergyTotalConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['Total'])


class TasmotaEnergyApparentPowerConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['ApparentPower'])


class TasmotaEnergyReactivePowerConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['ReactivePower'])


class TasmotaEnergyFactorConnector(AbstractTasmotaConnector[float]):
    type = float

    def __init__(self, _interface: TasmotaInterface):
        super().__init__("ENERGY")

    def _decode(self, value: JSONType) -> float:
        assert isinstance(value, dict)
        return float(value['Factor'])


class TasmotaOnlineConnector(Readable[bool], Subscribable[bool]):
    type = bool

    def __init__(self):
        super().__init__()
        self.value = False

    def _update_from_mqtt(self, msg: MQTTMessage) -> None:
        self.value = msg.payload == b'Online'
        self._publish(self.value, [])

    async def read(self) -> bool:
        return self.value



# Copyright 2021 Michael Thies <mail@mhthies.de>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import abc
import asyncio
import logging
from collections import defaultdict, deque
from typing import (
    NamedTuple, List, Optional, Generic, Type, Any, DefaultDict, Tuple, Callable, Deque, Set, TYPE_CHECKING)
import ctypes as c

if TYPE_CHECKING:
    from pulsectl import (
        PulseEventInfo, PulseSinkInfo, PulseSourceInfo, PulseServerInfo, PulseVolumeInfo)
    from pulsectl_asyncio import PulseAsync

import shc.conversion
from shc.base import Connectable, Subscribable, Readable, T, UninitializedError, Writable
from shc.datatypes import RangeFloat1, Balance
from shc.interfaces._helper import SupervisedClientInterface

logger = logging.getLogger(__name__)


class PulseVolumeRaw(NamedTuple):
    """
    "raw" representation of the volume setting of a Pulseaudio sink or source with individual channel volume values.

    :param values: List of float values of the volume setting of each individual channel of the sink/source from 0.0 to
            1.0
    :param map: Pulseaudio channel map as a list of integer channel positions, according to the libpulse
            `pa_channel_position` enum. E.g. [0] for mono; [1, 2] for normal stereo (left, right); [1, 2, 5, 6, 3, 7]
            for typical 5.1 surround devices. See https://freedesktop.org/software/pulseaudio/doxygen/channelmap_8h.html
            for reference. This is required for converting into the component-based :class:`PulseVolumeComponents`
            representation.
    """
    values: list = []
    map: list = []


class PulseVolumeComponents(NamedTuple):
    """
    abstract "component-based" representation of the volume setting of a Pulseaudio sink or source

    .. warning:
        Conversion methods :meth:`as_channels` and :meth:`from_channels`, which are used as default converters into/from
        :class:`PulseVolumeRaw` objects, require the libpulse shared library. This might be a problem when using and
        converting these datatypes on a remote SHC server, different from the SHC instance running the
        :class:`PulseAudioInterface`.

    :param volume: Master volume of the sink/source. Corresponds to the maximum individual channel volume
    :param balance: left/right balance (-1.0 means 100% left, +1.0 means 100% right).
    :param fade: rear/front balance (-1.0 means 100% rear speakers, +1.0 means 100% front speakers). If no surround
            channels are present, it is always 0.0.
    :param lfe_balance: subwoofer balance (-1.0 means no subwoofer, +1.0 means only subwoofer). If no subwoofer/LFE
            channel is present, the value is always 0.0
    :param normalized_values: The channel values after resetting the master volume, balance, fade and lfe_balance
            to their default values. This allows to modify these volume components but keep additional individual
            channel differences intact.
    :param map: Pulseaudio channel map as a list of integer channel positions, according to the libpulse
            `pa_channel_position` enum. E.g. [0] for mono; [1, 2] for normal stereo (left, right); [1, 2, 5, 6, 3, 7]
            for typical 5.1 surround devices. See https://freedesktop.org/software/pulseaudio/doxygen/channelmap_8h.html
            for reference. This is required for the conversion functions to/from :class:`PulseVolumeRaw`.
    """
    volume: RangeFloat1 = RangeFloat1(1.0)
    balance: Balance = Balance(0.0)
    fade: Balance = Balance(0.0)
    lfe_balance: Balance = Balance(0.0)
    normalized_values: list = []
    map: list = []

    def as_channels(self) -> PulseVolumeRaw:
        """
        Assemble a :class:`PulseVolumeRaw` volume object from the component-based representation and the additional
        derivations in the `normalized_values` field.

        This method uses C function bindings from libpulse to apply the balance, face, etc. to the channel
        volume values, using the channel map. Thus, the `libpulse` C shared library is required to use this method.

        This method is the SHC default converter from :class:`PulseVolumeComponents` to :class:`PulseVolumeRaw`.
        """
        from pulsectl._pulsectl import PA_CVOLUME, PA_CHANNEL_MAP, PA_CHANNELS_MAX, PA_VOLUME_NORM
        from shc.interfaces._pulse_ffi import pa_volume_t, pa_cvolume_set_lfe_balance, pa_cvolume_set_fade, \
            pa_cvolume_set_balance, pa_cvolume_scale

        num_channels = len(self.normalized_values)
        cvolume = PA_CVOLUME(num_channels, (pa_volume_t * PA_CHANNELS_MAX)(*self.normalized_values))
        cmap = PA_CHANNEL_MAP(num_channels, (c.c_int * PA_CHANNELS_MAX)(*self.map))
        pa_cvolume_set_lfe_balance(cvolume, cmap, self.lfe_balance)
        pa_cvolume_set_fade(cvolume, cmap, self.fade)
        pa_cvolume_set_balance(cvolume, cmap, self.balance)
        pa_cvolume_scale(cvolume, round(self.volume * PA_VOLUME_NORM))
        return PulseVolumeRaw([v / PA_VOLUME_NORM for v in cvolume.values[:cvolume.channels]], self.map)

    @classmethod
    def from_channels(cls, raw_volume: PulseVolumeRaw) -> "PulseVolumeComponents":
        """
        Convert a :class:`PulseVolumeRaw` volume object into the component-based representation.

        This method uses C function bindings from libpulse to calculate the balance, face, etc. from the channel volume
        values and the channel map. Thus, the `libpulse` C shared library is required to use this method.

        This method is the SHC default converter from :class:`PulseVolumeRaw` to :class:`PulseVolumeComponents`.
        """
        from pulsectl._pulsectl import PA_CVOLUME, PA_CHANNEL_MAP, PA_CHANNELS_MAX, PA_VOLUME_NORM
        from shc.interfaces._pulse_ffi import pa_volume_t, pa_cvolume_set_lfe_balance, pa_cvolume_set_fade, \
            pa_cvolume_set_balance, pa_cvolume_scale, pa_cvolume_max, pa_cvolume_get_balance, pa_cvolume_get_fade, \
            pa_cvolume_get_lfe_balance

        num_channels = len(raw_volume.values)
        cvolume = PA_CVOLUME(num_channels, (pa_volume_t * PA_CHANNELS_MAX)(*(round(PA_VOLUME_NORM * v)
                                                                             for v in raw_volume.values)))
        cmap = PA_CHANNEL_MAP(num_channels, (c.c_int * PA_CHANNELS_MAX)(*raw_volume.map))
        volume = pa_cvolume_max(cvolume) / PA_VOLUME_NORM
        pa_cvolume_scale(cvolume, PA_VOLUME_NORM)
        balance = pa_cvolume_get_balance(cvolume, cmap)
        pa_cvolume_set_balance(cvolume, cmap, 0.0)
        fade = pa_cvolume_get_fade(cvolume, cmap)
        pa_cvolume_set_fade(cvolume, cmap, 0.0)
        lfe_balance = pa_cvolume_get_lfe_balance(cvolume, cmap)
        pa_cvolume_set_lfe_balance(cvolume, cmap, 0.0)
        return cls(RangeFloat1(volume), Balance(balance), Balance(fade), Balance(lfe_balance),
                   list(cvolume.values)[:num_channels], raw_volume.map)


shc.conversion.register_converter(PulseVolumeRaw, PulseVolumeComponents, PulseVolumeComponents.from_channels)
shc.conversion.register_converter(PulseVolumeComponents, PulseVolumeRaw, lambda v: v.as_channels())


class PulseAudioInterface(SupervisedClientInterface):
    """
    Interface for controlling a Pulseaudio server and receiving status feedback, esp. sink and source volumes, muting,
    and level monitoring.

    The interface is based on the pulsectl and pulsectl-asyncio python packages, which internally use C function
    bindings to the Pulseaudio client shared library `libpulse`. Thus, this library must be installed on the system
    where SHC is running, for the PulseAudio interface to work. In Debian and Ubuntu, its contained in the `libpulse0`
    package, on Arch Linux the package is named `libpulse`.

    Currently the interface provides connectors for the following tasks:

    - subscribe to, read and change volume of a Pulseaudio sink (output device) or source (input device)
    - subscribe to, read and change mute state of a Pulseaudio sink or source
    - subscribe to and read state (running vs. idle/suspended) of a sink or source
    - monitor audio level of a sink or source
    - subscribe to, read and change the current default sink or default source by name

    All of the sink/source connectors are available in two different flavors: They can either be bound to a specific
    sink/source by name or follow the current default sink/source of the Pulseaudio server. Reading from connectors with
    a fixed sink/source name will raise an :class`shc.UninitializedError` if no such sink/source is present. They will
    automatically become active as soon as the sink/source appears. Default sink/source connectors will always publish
    the new respective value when the default sink/source changes on the server.

    Named sink/source connectors:

    - :meth:`sink_volume`
    - :meth:`sink_muted`
    - :meth:`sink_running`
    - :meth:`sink_peak_monitor`
    - :meth:`source_volume`
    - :meth:`source_muted`
    - :meth:`source_running`
    - :meth:`source_peak_monitor`

    Default sink/source connectors:

    - :meth:`default_sink_volume`
    - :meth:`default_sink_muted`
    - :meth:`default_sink_running`
    - :meth:`default_sink_peak_monitor`
    - :meth:`default_source_volume`
    - :meth:`default_source_muted`
    - :meth:`default_source_running`
    - :meth:`default_source_peak_monitor`

    Server state connectors:

    - :meth:`default_sink_name`
    - :meth:`default_source_name`

    All \\*_volume connectors use the :class:`PulseVolumeRaw` type for representing the sink's/source's volume setting,
    which contains a float volume setting for each individual channel of the sink/source (e.g. 6 individual values for a
    5.1 surround codec). Typically, you'll want to show and control the volume in more tangible components like
    *master volume*, *balance*, *fade* (front/rear) and subwoofer balance. For this, the raw volume can be converted
    into the :class:`PulseVolumeComponents` type, representing these components as separate fields (but still keeping
    the information about any further derivations in the individual channels). Thus, a typical application of the
    volume connectors would look like this::

        interface = PulseAudioInterface()
        volume_components = shc.Variable(PulseVolumeComponents)\\
            .connect(interface.default_sink_volume(), convert=True)  # convert PulseVolumeRaw and PulseVolumeComponents

        slider = shc.web.widgets.Slider("Volume").connect(volume_components.field('volume'))
        balance_slider = shc.web.widgets.Slider("Balance").connect(volume_components.field('balance'), convert=True)


    :param pulse_client_name: Client name reported to the Pulseaudio server when connecting.
    :param pulse_server_socket: Address of the Pulseaudio server socket, e.g. "unix:/run/user/1000/pulse/native". If not
        specified or None, the system default Pulseaudio instance is used.
    :param auto_reconnect: If True (default), the interface tries to reconnect automatically with
        exponential backoff (1.25 ^ n seconds sleep), when the Pulseaudio event subscription exits unexpectedly, e.g.
        due to a connection loss with the Pulseaudio server. Otherwise, the complete SHC system is shut down on
        such errors.
    :param failsafe_start: If True and auto_reconnect is True, the interface allows SHC to start up, even if the
        connection and event subscription with the Pulseaudio server fails in the first try. The connection is retried
        in background with exponential backoff (see `auto_reconnect` option). Otherwise (default), the first connection
        attempt on startup is not retried and will raise an exception from `start()` on failure, even if
        `auto_reconnect` is True.
    """
    def __init__(self, pulse_client_name: str = "smarthomeconnect", pulse_server_socket: Optional[str] = None,
                 auto_reconnect: bool = True, failsafe_start: bool = False):
        super().__init__(auto_reconnect, failsafe_start)
        from pulsectl_asyncio import PulseAsync

        self.pulse = PulseAsync(pulse_client_name, server=pulse_server_socket)
        self.sink_connectors_by_id: DefaultDict[int, List["SinkConnector"]] = defaultdict(list)
        self.sink_connectors_by_name: DefaultDict[Optional[str], List["SinkConnector"]] = defaultdict(list)
        self.source_connectors_by_id: DefaultDict[int, List["SourceConnector"]] = defaultdict(list)
        self.source_connectors_by_name: DefaultDict[Optional[str], List["SourceConnector"]] = defaultdict(list)

        self._default_sink_name_connector = DefaultNameConnector(
            self.pulse, "default_sink_name", self._register_origin_callback)
        self._default_source_name_connector = DefaultNameConnector(
            self.pulse, "default_source_name", self._register_origin_callback)
        #: A subset of {'source', 'sink', 'server'}
        self._subscribe_facilities: Set[str] = set()

        #: A dict for keeping track of the SHC value update origin of changes we are currently applying to the
        #: Pulseaudio server, so that we can correctly apply the original origin list to the resulting Pulseaudio event
        self._change_event_origin: DefaultDict[Tuple[str, int], Deque[List[Any]]] = defaultdict(deque)

    async def _connect(self) -> None:
        await self.pulse.connect()

    async def _disconnect(self) -> None:
        self.pulse.disconnect()

    async def _subscribe(self) -> None:
        server_info = await self.pulse.server_info()
        self._default_sink_name_connector._update(server_info, [])
        self._default_source_name_connector._update(server_info, [])
        for sink_info in await self.pulse.sink_list():
            sink_connectors = self.sink_connectors_by_name.get(sink_info.name, [])
            if sink_info.name == server_info.default_sink_name:
                sink_connectors += self.sink_connectors_by_name.get(None, [])
            if sink_connectors:
                logging.debug("Pulseaudio sink \"%s\" is already available. Mapping %s connectors to sink id %s",
                              sink_info.name, len(sink_connectors), sink_info.index)
            for connector in sink_connectors:
                try:
                    connector.change_id(sink_info.index)
                    self.sink_connectors_by_id[sink_info.index].append(connector)
                    connector.on_change(sink_info, [])
                except Exception as e:
                    logging.error("Error while initializing connector %s with current data of sink", connector,
                                  exc_info=e)
        for source_info in await self.pulse.source_list():
            source_connectors = self.source_connectors_by_name.get(source_info.name, [])
            if source_info.name == server_info.default_source_name:
                source_connectors += self.source_connectors_by_name.get(None, [])
            if source_connectors:
                logging.debug("Pulseaudio source \"%s\" is already available. Mapping %s connectors to source id %s",
                              source_info.name, len(source_connectors), source_info.index)
            for source_connector in source_connectors:
                try:
                    source_connector.change_id(source_info.index)
                    self.source_connectors_by_id[source_info.index].append(source_connector)
                    source_connector.on_change(source_info, [])
                except Exception as e:
                    logging.error("Error while initializing connector %s with current data of source", source_connector,
                                  exc_info=e)

    async def _run(self) -> None:
        self._running.set()
        async for event in self.pulse.subscribe_events(*self._subscribe_facilities):
            try:
                await self._dispatch_pulse_event(event)
            except Exception as e:
                logger.error("Error while dispatching PulseAudio event %s", event, exc_info=e)

    async def _dispatch_pulse_event(self, event: "PulseEventInfo") -> None:
        from pulsectl import PulseEventTypeEnum, PulseEventFacilityEnum
        logger.debug("Dispatching Pulse audio event: %s", event)

        if event.t is PulseEventTypeEnum.new:
            if event.facility is PulseEventFacilityEnum.sink:
                data = await self.pulse.sink_info(event.index)
                name = data.name
                for connector in self.sink_connectors_by_name.get(name, []):
                    connector.change_id(event.index)
                    self.sink_connectors_by_id[event.index].append(connector)
                    connector.on_change(data, [])
            elif event.facility is PulseEventFacilityEnum.source:
                data = await self.pulse.source_info(event.index)
                name = data.name
                for source_connector in self.source_connectors_by_name.get(name, []):
                    source_connector.change_id(event.index)
                    self.source_connectors_by_id[event.index].append(source_connector)
                    source_connector.on_change(data, [])

        elif event.t is PulseEventTypeEnum.remove:
            if event.facility is PulseEventFacilityEnum.sink:
                for connector in self.sink_connectors_by_id.get(event.index, []):
                    connector.change_id(None)
                if event.index in self.sink_connectors_by_id:
                    del self.sink_connectors_by_id[event.index]
            elif event.facility is PulseEventFacilityEnum.source:
                for source_connector in self.source_connectors_by_id.get(event.index, []):
                    source_connector.change_id(None)
                if event.index in self.source_connectors_by_id:
                    del self.source_connectors_by_id[event.index]

        elif event.t is PulseEventTypeEnum.change:
            # Check if the event has probably been caused by a value update from SHC. In this case, we should have the
            # original origin stored in self.event_origin and can apply it to the publishing of the event
            try:
                index = 0 if event.facility == PulseEventFacilityEnum.server else event.index
                origin = self._change_event_origin[(event.facility._value, index)].popleft()
            except IndexError:
                origin = []

            if event.facility is PulseEventFacilityEnum.sink:
                data = await self.pulse.sink_info(event.index)
                for connector in self.sink_connectors_by_id.get(event.index, []):
                    connector.on_change(data, origin)
            elif event.facility is PulseEventFacilityEnum.source:
                data = await self.pulse.source_info(event.index)
                for source_connector in self.source_connectors_by_id.get(event.index, []):
                    source_connector.on_change(data, origin)

            elif event.facility is PulseEventFacilityEnum.server:
                # For server change events, we need to update our default_*_name connectors and possibly change the
                # current_id of all the default_sink/source_* connectors (and update them with the current state of the
                # new default sink/source)
                server_info = await self.pulse.server_info()
                self._default_sink_name_connector._update(server_info, origin)
                self._default_source_name_connector._update(server_info, origin)
                # Update default sink/source connectors
                default_sink_data = await self.pulse.get_sink_by_name(server_info.default_sink_name)
                default_source_data = await self.pulse.get_source_by_name(server_info.default_source_name)
                for connector in self.sink_connectors_by_name.get(None, []):
                    if connector.current_id is not None:
                        try:
                            self.sink_connectors_by_id[connector.current_id].remove(connector)
                        except ValueError:
                            pass
                    connector.change_id(default_sink_data.index)
                    self.sink_connectors_by_id[default_sink_data.index].append(connector)
                    connector.on_change(default_sink_data, [])  # should we set the original origin here?
                for source_connector in self.source_connectors_by_name.get(None, []):
                    if source_connector.current_id is not None:
                        try:
                            self.source_connectors_by_id[source_connector.current_id].remove(source_connector)
                        except ValueError:
                            pass
                    source_connector.change_id(default_source_data.index)
                    self.source_connectors_by_id[default_source_data.index].append(source_connector)
                    source_connector.on_change(default_source_data, [])  # should we set the original origin here?

    def _register_origin_callback(self, facility: str, index: int, origin: List[Any]) -> None:
        # Avoid infinite growing of event_origin dict's lists, by adding origins of events that will never be
        # removed b/c we do not subscribe this kind of events.
        if facility not in self._subscribe_facilities:
            return
        self._change_event_origin[(facility, index)].append(origin)

    def sink_volume(self, sink_name: str) -> "SinkVolumeConnector":
        self._subscribe_facilities.add('sink')
        connector = SinkVolumeConnector(self.pulse, self._register_origin_callback)
        self.sink_connectors_by_name[sink_name].append(connector)
        return connector

    def sink_muted(self, sink_name: str) -> "SinkMuteConnector":
        self._subscribe_facilities.add('sink')
        connector = SinkMuteConnector(self.pulse, self._register_origin_callback)
        self.sink_connectors_by_name[sink_name].append(connector)
        return connector

    def sink_running(self, sink_name: str) -> "SinkStateConnector":
        self._subscribe_facilities.add('sink')
        connector = SinkStateConnector(self.pulse)
        self.sink_connectors_by_name[sink_name].append(connector)
        return connector

    def sink_peak_monitor(self, sink_name: str, frequency: int = 1) -> "SinkPeakConnector":
        self._subscribe_facilities.add('sink')
        connector = SinkPeakConnector(self.pulse, frequency)
        self.sink_connectors_by_name[sink_name].append(connector)
        return connector

    def default_sink_volume(self) -> "SinkVolumeConnector":
        self._subscribe_facilities.add('server')
        self._subscribe_facilities.add('sink')
        connector = SinkVolumeConnector(self.pulse, self._register_origin_callback)
        self.sink_connectors_by_name[None].append(connector)
        return connector

    def default_sink_muted(self) -> "SinkMuteConnector":
        self._subscribe_facilities.add('server')
        self._subscribe_facilities.add('sink')
        connector = SinkMuteConnector(self.pulse, self._register_origin_callback)
        self.sink_connectors_by_name[None].append(connector)
        return connector

    def default_sink_running(self) -> "SinkStateConnector":
        self._subscribe_facilities.add('server')
        self._subscribe_facilities.add('sink')
        connector = SinkStateConnector(self.pulse)
        self.sink_connectors_by_name[None].append(connector)
        return connector

    def default_sink_peak_monitor(self, frequency: int = 1) -> "SinkPeakConnector":
        self._subscribe_facilities.add('server')
        connector = SinkPeakConnector(self.pulse, frequency)
        self.sink_connectors_by_name[None].append(connector)
        return connector

    def default_sink_name(self) -> "DefaultNameConnector":
        self._subscribe_facilities.add('server')
        return self._default_sink_name_connector

    def source_volume(self, source_name: str) -> "SourceVolumeConnector":
        self._subscribe_facilities.add('source')
        connector = SourceVolumeConnector(self.pulse, self._register_origin_callback)
        self.source_connectors_by_name[source_name].append(connector)
        return connector

    def source_muted(self, source_name: str) -> "SourceMuteConnector":
        self._subscribe_facilities.add('source')
        connector = SourceMuteConnector(self.pulse, self._register_origin_callback)
        self.source_connectors_by_name[source_name].append(connector)
        return connector

    def source_running(self, source_name: str) -> "SourceStateConnector":
        self._subscribe_facilities.add('source')
        connector = SourceStateConnector(self.pulse)
        self.source_connectors_by_name[source_name].append(connector)
        return connector

    def source_peak_monitor(self, source_name: str, frequency: int = 1) -> "SourcePeakConnector":
        self._subscribe_facilities.add('source')
        connector = SourcePeakConnector(self.pulse, frequency)
        self.source_connectors_by_name[source_name].append(connector)
        return connector

    def default_source_volume(self) -> "SourceVolumeConnector":
        self._subscribe_facilities.add('source')
        self._subscribe_facilities.add('server')
        connector = SourceVolumeConnector(self.pulse, self._register_origin_callback)
        self.source_connectors_by_name[None].append(connector)
        return connector

    def default_source_muted(self) -> "SourceMuteConnector":
        self._subscribe_facilities.add('source')
        self._subscribe_facilities.add('server')
        connector = SourceMuteConnector(self.pulse, self._register_origin_callback)
        self.source_connectors_by_name[None].append(connector)
        return connector

    def default_source_running(self) -> "SourceStateConnector":
        self._subscribe_facilities.add('source')
        self._subscribe_facilities.add('server')
        connector = SourceStateConnector(self.pulse)
        self.source_connectors_by_name[None].append(connector)
        return connector

    def default_source_peak_monitor(self, frequency: int = 1) -> "SourcePeakConnector":
        self._subscribe_facilities.add('server')
        connector = SourcePeakConnector(self.pulse, frequency)
        self.source_connectors_by_name[None].append(connector)
        return connector

    def default_source_name(self) -> "DefaultNameConnector":
        self._subscribe_facilities.add('server')
        return self._default_source_name_connector


class SinkConnector(metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()
        self.current_id: Optional[int] = None

    def on_change(self, data: "PulseSinkInfo", origin: List[Any]) -> None:
        pass

    def change_id(self, index: Optional[int]) -> None:
        self.current_id = index


class SourceConnector(metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()
        self.current_id: Optional[int] = None

    def on_change(self, data: "PulseSourceInfo", origin: List[Any]) -> None:
        pass

    def change_id(self, index: Optional[int]) -> None:
        self.current_id = index


class SinkAttributeConnector(Subscribable[T], Readable[T], SinkConnector, Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, pulse: "PulseAsync", type_: Type[T]):
        self.type = type_
        super().__init__()
        self.pulse = pulse

    def on_change(self, data: "PulseSinkInfo", origin: List[Any]) -> None:
        self._publish(self._convert_from_pulse(data), origin)

    @abc.abstractmethod
    def _convert_from_pulse(self, data: "PulseSinkInfo") -> T:
        pass

    async def read(self) -> T:
        if self.current_id is None:
            raise UninitializedError()
        data = await self.pulse.sink_info(self.current_id)
        return self._convert_from_pulse(data)


class SourceAttributeConnector(Subscribable[T], Readable[T], SourceConnector, Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, pulse: "PulseAsync", type_: Type[T]):
        self.type = type_
        super().__init__()
        self.pulse = pulse

    def on_change(self, data: "PulseSourceInfo", origin: List[Any]) -> None:
        self._publish(self._convert_from_pulse(data), origin)

    @abc.abstractmethod
    def _convert_from_pulse(self, data: "PulseSourceInfo") -> T:
        pass

    async def read(self) -> T:
        if self.current_id is None:
            raise UninitializedError()
        data = await self.pulse.source_info(self.current_id)
        return self._convert_from_pulse(data)


class SinkStateConnector(SinkAttributeConnector[bool]):
    def __init__(self, pulse: "PulseAsync"):
        super().__init__(pulse, bool)

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> bool:
        from pulsectl import PulseStateEnum

        return data.state == PulseStateEnum.running

    def change_id(self, index: Optional[int]) -> None:
        super().change_id(index)
        if index is None:
            self._publish(False, [])


class SourceStateConnector(SourceAttributeConnector[bool]):
    def __init__(self, pulse: "PulseAsync"):
        super().__init__(pulse, bool)

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> bool:
        from pulsectl import PulseStateEnum

        return data.state == PulseStateEnum.running

    def change_id(self, index: Optional[int]) -> None:
        super().change_id(index)
        if index is None:
            self._publish(False, [])


class SinkMuteConnector(SinkAttributeConnector[bool], Writable[bool]):
    def __init__(self, pulse: "PulseAsync", register_origin_callback: Callable[[str, int, List[Any]], None]):
        super().__init__(pulse, bool)
        self.register_origin_callback = register_origin_callback

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> bool:
        return bool(data.mute)

    async def _write(self, value: T, origin: List[Any]) -> None:
        if self.current_id is None:
            raise RuntimeError("PulseAudio sink id for {} is currently not defined".format(repr(self)))
        self.register_origin_callback('sink', self.current_id, origin)
        await self.pulse.sink_mute(self.current_id, value)


class SourceMuteConnector(SourceAttributeConnector[bool], Writable[bool]):
    def __init__(self, pulse: "PulseAsync", register_origin_callback: Callable[[str, int, List[Any]], None]):
        super().__init__(pulse, bool)
        self.register_origin_callback = register_origin_callback

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> bool:
        return bool(data.mute)

    async def _write(self, value: T, origin: List[Any]) -> None:
        if self.current_id is None:
            raise RuntimeError("PulseAudio source id for {} is currently not defined".format(repr(self)))
        self.register_origin_callback('source', self.current_id, origin)
        await self.pulse.source_mute(self.current_id, value)


class SinkVolumeConnector(SinkAttributeConnector[PulseVolumeRaw], Writable[PulseVolumeRaw]):
    def __init__(self, pulse: "PulseAsync", register_origin_callback: Callable[[str, int, List[Any]], None]):
        super().__init__(pulse, PulseVolumeRaw)
        self.register_origin_callback = register_origin_callback

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> PulseVolumeRaw:
        return PulseVolumeRaw(data.volume.values, data.channel_list_raw)

    async def _write(self, value: PulseVolumeRaw, origin: List[Any]) -> None:
        from pulsectl import PulseVolumeInfo

        if self.current_id is None:
            raise RuntimeError("PulseAudio sink id for {} is currently not defined".format(repr(self)))
        self.register_origin_callback('sink', self.current_id, origin)
        await self.pulse.sink_volume_set(self.current_id, PulseVolumeInfo(value.values))


class SourceVolumeConnector(SourceAttributeConnector[PulseVolumeRaw], Writable[PulseVolumeRaw]):
    def __init__(self, pulse: "PulseAsync", register_origin_callback: Callable[[str, int, List[Any]], None]):
        super().__init__(pulse, PulseVolumeRaw)
        self.register_origin_callback = register_origin_callback

    def _convert_from_pulse(self, data: "PulseSinkInfo") -> PulseVolumeRaw:
        return PulseVolumeRaw(data.volume.values, data.channel_list_raw)

    async def _write(self, value: PulseVolumeRaw, origin: List[Any]) -> None:
        from pulsectl import PulseVolumeInfo

        if self.current_id is None:
            raise RuntimeError("PulseAudio source id for {} is currently not defined".format(repr(self)))
        self.register_origin_callback('source', self.current_id, origin)
        await self.pulse.source_volume_set(self.current_id, PulseVolumeInfo(value.values))


class SinkPeakConnector(SinkConnector, Subscribable[RangeFloat1]):
    type = RangeFloat1

    def __init__(self, pulse: "PulseAsync", frequency: int):
        super().__init__()
        self.pulse = pulse
        self.frequency = frequency
        self.task: Optional[asyncio.Task] = None

    def change_id(self, index: Optional[int]) -> None:
        if self.task and index != self.current_id:
            self.task.cancel()
            self.task = None
        super().change_id(index)
        if index is not None and not self.task:
            self.task = asyncio.create_task(self._run())

    async def _run(self) -> None:
        from pulsectl import PulseDisconnected

        data = await self.pulse.sink_info(self.current_id)
        try:
            async for value in self.pulse.subscribe_peak_sample(data.monitor_source_name, self.frequency):
                self._publish(RangeFloat1(value), [])
        except (asyncio.CancelledError, PulseDisconnected):
            pass
        except Exception as e:
            logger.error("Error while monitoring peaks of sink %s:", data.name, exc_info=e)


class SourcePeakConnector(SourceConnector, Subscribable[RangeFloat1]):
    type = RangeFloat1

    def __init__(self, pulse: "PulseAsync", frequency: int):
        super().__init__()
        self.pulse = pulse
        self.frequency = frequency
        self.task: Optional[asyncio.Task] = None

    def change_id(self, index: Optional[int]) -> None:
        if self.task and index != self.current_id:
            self.task.cancel()
            self.task = None
        super().change_id(index)
        if index is not None and not self.task:
            self.task = asyncio.create_task(self._run())

    async def _run(self) -> None:
        from pulsectl import PulseDisconnected

        data = await self.pulse.source_info(self.current_id)
        try:
            async for value in self.pulse.subscribe_peak_sample(data.name, self.frequency):
                self._publish(RangeFloat1(value), [])
        except (asyncio.CancelledError, PulseDisconnected):
            pass
        except Exception as e:
            logger.error("Error while monitoring peaks of source %s:", data.name, exc_info=e)


class DefaultNameConnector(Subscribable[str], Readable[str], Writable[str]):
    type = str

    def __init__(self, pulse: "PulseAsync", attr: str, register_origin_callback: Callable[[str, int, List[Any]], None]):
        super().__init__()
        self.pulse = pulse
        self.attr = attr
        self.register_origin_callback = register_origin_callback

    def _update(self, server_info: "PulseServerInfo", origin: List[Any]) -> None:
        self._publish(getattr(server_info, self.attr), origin)

    async def read(self) -> str:
        server_info = await self.pulse.server_info()
        return getattr(server_info, self.attr)

    async def _write(self, value: str, origin: List[Any]) -> None:
        self.register_origin_callback('server', 0, origin)
        if self.attr == 'default_source_name':
            await self.pulse.source_default_set(value)
        else:
            await self.pulse.sink_default_set(value)
